################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
align_pairs.py - align nucleotide sequence pairs
================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

Read a file of unaligned nucleotide sequence pairs and align them.

Usage
-----

Example::

   python align_pairs.py --help

Type::

   python align_pairs.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import getopt
import tempfile
import optparse
import CGAT.Experiment as E
import CGAT.AlignedPairs as AlignedPairs
import CGAT.Blat as Blat
import CGAT.FastaIterator as FastaIterator
import CGAT.IndexedFasta as IndexedFasta
import CGAT.IOTools as IOTools

def iterate_double_fasta ( fn1, fn2 ):
    iterator = FastaIterator.iterate_together( fn1, fn2 )
    for seq1, seq2 in iterator:
        yield AlignedPairs.UnalignedPair( 
            token1 = seq1.title,
            sequence1 = seq1.sequence,
            token2 = seq2.title,
            sequence2 = seq2.sequence )

def iterate_single_fasta ( fn1 ):
    iterator = FastaIterator.FastaIterator( fn1 )
    while 1:
        seq1, seq2 = iterator.next(), iterator.next()
        yield AlignedPairs.UnalignedPair( 
            token1 = seq1.title,
            sequence1 = seq1.sequence,
            token2 = seq2.title,
            sequence2 = seq2.sequence )

def iterate_list ( infile, idx1, idx2 = None):

    fasta1 = IndexedFasta.IndexedFasta( idx1 )
    if idx2 == None:
        fasta2 = fasta1
    else:
        fasta2 = IndexedFasta.IndexedFasta( idx2 )

    first = True
    for line in infile:
        if line[0] == "#": continue
        id1, id2 = line[:-1].split("\t")[:2]

        try:
            yield AlignedPairs.UnalignedPair( 
                token1 = id1,
                sequence1 = fasta1.getSequence( id1 ),
                token2 = id2,
                sequence2 = fasta2.getSequence( id2 ) )
        except KeyError, msg:
            if first:
                first = False
                continue
            raise KeyError(msg)
    
def iterate_table( infile ):
    for line in infile:
        if line[0] == "#": continue
        
        unaligned_pair = AlignedPairs.UnalignedPair()
        unaligned_pair.Read(line)
        yield unaligned_pair


def getFile( section, options ):
    if options.output_filename_pattern:
        if "%s" in options.output_filename_pattern:
            return open( options.output_filename_pattern % section, "w" )
        else:
            return open( options.output_filename_pattern, "w" )
    return options.stdout
        
##------------------------------------------------------------
if __name__ == '__main__':

    parser = E.OptionParser( version = "%prog version: $Id: align_pairs.py 2781 2009-09-10 11:33:14Z andreas $", usage = globals()["__doc__"] )

    parser.add_option("--skip-statistics", dest="skip_stats", action="store_true",
                      help="do not compute alignment statistics [%default]."  )

    parser.add_option("--method", dest="methods", type="choice", action="append",
                      choices=("dialign", "clustal", "blastz", "nw", "sw", "dba", "dialignlgs" ),
                      help="alignment method [%default]."  )

    parser.add_option("--anchor-alignment", dest="anchor_alignment", type="int",
                      help="anchor alignmet with xxx residues [%default]."  )

    parser.add_option("--output-format", dest="output_formats", type="choice", action="append",
                      choices=("fasta", "stats", "psl" ),
                      help="anchor alignment with xxx residues [%default]."  )

    parser.add_option("--input-format", dest="input_format", type="choice", 
                      choices=("fasta", "list" ),
                      help="input format of stdin [%default]."  )

    parser.add_option("--output-filename-pattern", dest="output_filename_pattern", type="string",
                      help="output pattern for multiple files [%default]."  )

    parser.add_option("--filename-sequences1", dest="filename_sequences1", type="string",
                      help="first indexed input filename with sequences [%default]."  )

    parser.add_option("--filename-sequences2", dest="filename_sequences2", type="string",
                      help="second indexed input filename with sequences [%default]."  )

    parser.add_option("--options-blastz", dest="options_blastz", type="string",
                      help="command line options for blastz [%default]."  )

    parser.set_defaults( 
        skip_stats = False,
        methods = [],
        output_formats = [],
        input_format = "fasta",
        output_filename_pattern = None,
        filename_sequences1 = None,
        filename_sequences2 = None,
        anchor_alignment = 0,
        options_blastz = "C=2 B=1 T=0 W=6 K=2200" )

    (options, args) = E.Start( parser, add_pipe_options = True )

    if len(options.methods) == 0:
        print USAGE
        print "please specify an alignment method."
        sys.exit(1)

    if len(options.output_formats) == 0:
        print USAGE
        print "please specify at least one output format."
        sys.exit(1)

    if len(args) == 2:
        iterator = iterate_double_fasta( args[0], args[1] )        
    elif options.filename_sequences1 and options.filename_sequences2:
        if len(args) == 0 or (len(args) == 1 and args[0] == "-"):
            infile = options.stdin
        elif len(args) == 1:
            infile = open( args[0], "r") 
                
        iterator = iterate_list( infile, options.filename_sequences1, options.filename_sequences2 )
    else:
        iterator = iterate_single_fasta( options.stdin )
        

    npairs, ntoken_pairs = 0, 0
    ninput, nskipped, nerrors = 0, 0, 0

    outfile_table = None
    outfile_fasta = None
    outfile_psl = None
    if "table" in options.output_formats:
        outfile_table = getFile( "table ", options )
        outfile_table.write( """# CATEGORY:       category [intron|exon]
# METHOD:         alignment method
# TOKEN:          name
# ID:             segment id
# TOTAL:          number of segments
# LEN:            length of segment
# NALIGNED:       number of aligned positions
# PALIGNED:       percentage of aligned positions
# IDENT:          number of identical positions
# TRANSIT:        number of transitions
# TRANSVERS:      number of transversion
# MATCHES:        number of matching positions
# PIDENT:         percentage of identical positions
# PTRANSIT:       precentage of transitions
# PTRANSVERS:     precentage of transversion
# BLOCKSIZES:     alignment, length of blocks
# GAPS:           gap sizes in sequence 1/2
CATEGORY\tMETHOD\tTOKEN1\tID1\tTOTAL1\tLEN1\tTOKEN2\tID2\tTOTAL2\tLEN2\tNALIGNED\tPALIGNED\tIDENT\tTRANSIT\tTRANSVER\tMATCHES\tPIDENT\tPTRANSVIT\tPTRANVER\tBLOCKSIZES\tGAPSIZES\tGAPSIZES\tTYPE1\tTYPE2\n""")

    if "fasta" in options.output_formats:
        outfile_fasta = getFile( "fasta", options )

    if "psl" in options.output_formats:
        outfile_psl = getFile( "psl", options )

    ## setup alignment objects
    for unaligned_pair in iterator:

        ninput += 1
        
        for method in options.methods:

            pair = AlignedPairs.AlignedPair( unaligned_pair )
            pair.mOptionsBlastZ = options.options_blastz

            try:
                pair.Align( method, anchor = options.anchor_alignment )
            except AlignedPairs.AlignmentError, msg:
                
                if options.loglevel >= 1:
                    options.stdlog.write( "# %s - %s: %s\n" % (msg, unaligned_pair.mToken1, unaligned_pair.mToken2))
                    if options.loglevel >= 2:
                        options.stdlog.write( "# input=%s\n" % (str(unaligned_pair)))

                nskipped += 1
                continue

            if outfile_table:
                outfile_table.write( str(pair) + "\n" )
            
            if outfile_fasta:
                outfile_fasta.write( ">%s\n%s\n>%s\n%s\n" % (pair.mToken1, pair.mAlignedSequence1, pair.mToken2, pair.mAlignedSequence2 ) )

            if outfile_psl:
                entry = Blat.Match()
                entry.mQueryId, entry.mSbjctId = pair.mToken1, pair.mToken2
                entry.strand = pair.strand
                entry.fromMap( pair.mAlignment )
                outfile_psl.write( str(entry) + "\n" )

            npairs += 1
     
    E.info( "input=%i, skipped=%i, nerrors=%i, transcripts=%i, pairs=%i" % (ninput, nskipped,
                                                                            nerrors,
                                                                            ntoken_pairs, npairs ) )
    E.Stop()
    sys.exit(0)




################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
split_genomic_fasta_file.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python split_genomic_fasta_file.py --help

Type::

   python split_genomic_fasta_file.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import optparse

import CGAT.Experiment as E

USAGE="""python %s [OPTIONS] [genomic_sequence] [ < genomic sequence]

Version: $Id: split_genomic_fasta_file.py 1225 2007-04-10 15:13:11Z andreas $

If the file contains multiple sequence entries, it is split
at entry boundaries. Currently, individual entries are not
split.

""" % sys.argv[0]

global_written2file = 0
global_outfile = None
global_file_id = 0

def maskSequence( sequence, regions, mask_char = "N" ):
    """mask sequence with regions."""
    nmasked = 0
    for start,end in regions:
        sequence[start:end] = mask_char * (end-start)
        nmasked += end-start
        
    return sequence, nmasked

def processSequence( key, description_rest, sequence, options, mask_regions = None ):

    global global_outfile
    global global_file_id
    global global_written2file

    written = 0

    lsequence = len(sequence)

    if mask_regions:
        sequence, nmasked = maskSequence( sequence, mask_regions[key] )
    else:
        nmasked = 0

    while written < lsequence:

        ## size of current chunk
        chunk_size = min( options.chunk_size - global_written2file, lsequence - written)

        first_res = max(0, written )
        last_res  = min(written + chunk_size + options.extend, lsequence )

        if options.loglevel >= 2:
            options.stdlog.write( "# file_id=%i, chunk_size=%i, global_written2file=%i, written=%i, lsequence=%i, first_res=%i, last_res=%i\n" % \
                                  (global_file_id, chunk_size, global_written2file, written, lsequence, first_res, last_res ) )
            

        ## check if we need to open a new sequence file:
        if global_outfile == None:
            global_file_id += 1
            global_outfile = open( options.output_pattern % global_file_id, "w")
                
        offset_positive_strand = first_res
        offset_negative_strand = lsequence - last_res

        if options.format == "fasta":        
            global_outfile.write(">%s_%i_%i %s\n%s\n" % (key,
                                                         offset_positive_strand, offset_negative_strand,                                                     
                                                         description_rest, sequence[first_res:last_res]))
        elif options.format == "ranges":
            global_outfile.write("%s\t%i\t%i\n" % (key, first_res, last_res ) )

        global_written2file += chunk_size
        
        ## if we have reached the chunksize: close file
        if global_written2file >= options.chunk_size:
            global_outfile.close()
            global_outfile = None
            global_written2file = 0

        written += chunk_size

    return len(sequence), nmasked

if __name__ == '__main__':


    parser = E.OptionParser( version = "%prog version: $Id: split_genomic_fasta_file.py 1225 2007-04-10 15:13:11Z andreas $")

    parser.add_option("-e", "--extend", dest="extend", type="int",
                      help="extend sequences by # residues."  )

    parser.add_option("-o", "--output-pattern", dest="output_pattern", type="string",
                      help="output pattern for the genomic files. Should contain a %i."  )

    parser.add_option("-s", "--sequence", dest="filename_sequence", type="string",
                      help="sequence filename."  )

    parser.add_option("-c", "--chunk-size", dest="chunk_size", type="int",
                      help="chunk size in nucleotides."  )

    parser.add_option("-f", "--format", dest="format", type="choice",
                      choices=("fasta", "ranges"),
                      help="output format options."  )

    parser.add_option("-m", "--mask-regions", dest="filename_mask_regions", type="string",
                      help="mask regions - parameter is a filename of regions with forward strand coordinates in the format 'contig <tab> start <tab> end'."  )

    parser.set_defaults(
        extend = 0,
        output_pattern = "genome_%5i.fasta",
        format = "fasta",
        filename_sequence = None,
        filename_mask_regions = None,
        mask_char = "N" )

    (options, args) = E.Start( parser, add_pipe_options = True )

    ## read segments to mask
    
    if options.filename_mask_regions:
        mask_regions = {}
        infile = open(options.filename_mask_regions, "r")
        nregions = 0
        for line in infile:
            if line[0] == "#": continue
            contig, start, end = line[:-1].split("\t")
            if contig not in mask_regions: mask_regions[contig] = []
            mask_regions[contig].append( (int(start),int(end) ) )
            nregions += 1

        if options.loglevel >= 1:
            options.stdlog.write("# masking %i regions on %i contigs\n" % (nregions, len(mask_regions)))
            
        for contig, regions in mask_regions.items():
            regions.sort()
            
        infile.close()
    else:
        mask_regions = None
        
    if options.filename_sequence:
        infile = open(options.filename_sequence, "r")
    else:
        infile = sys.stdin

    ## For large genomic sequences, reading the whole data is
    ## a stretch for memory resources. Thus proceed sequence by
    ## sequence

    # read complete genomic sequences
    fragments = []
    sequences = []
    key = None
    description_rest = None
    total_keys, total_written, total_masked = 0, 0, 0
    
    for line in infile:
        if line[0] == "#": continue
        if line[0] == ">":
            if fragments:
                nwritten, nmasked = processSequence( key, description_rest, string.join(fragments, ""), options, mask_regions)
                total_written += nwritten
                total_masked += nmasked
                total_keys += 1
                
            fragments = []
            
            x = re.split( "\s", line[1:-1])
            key = x[0]
            description_rest = string.join(x[1:], " ")
            continue
        fragments.append( line[:-1] )
        
    nwritten, nmasked = processSequence( key, description_rest, string.join(fragments, ""), options, mask_regions)        

    total_written += nwritten
    total_masked += nmasked
    total_keys += 1
    
    if options.filename_sequence: infile.close()

    if options.loglevel >= 1:
        options.stdlog.write( "# nkeys=%i, nwritten=%i, nmasked=%i\n" % (total_keys, total_written, total_masked))

    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
mali2kaks.py - rate analysis with PAML
======================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script will do pairwise rate analysis on a set of multiply aligned
sequences in :term:`fasta` format.

PAML bails if sequences in a multiple alignment are not overlapping. Thus this
code checks for each pair the overlap. If a pair is not overlapping, it goes
into pairwise mode.

Usage
-----

Example::

   python mali2kaks.py --help

Type::

   python mali2kaks.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import tempfile
import subprocess
import optparse
import time
import math

from types import *

import CGAT.Experiment as E
import CGAT.WrapperCodeML as WrapperCodeML
import CGAT.Mali as Mali
import CGAT.Genomics as Genomics
import CGAT.RateEstimation as RateEstimation
import CGAT.IOTools as IOTools
import CGAT.TreeTools as TreeTools

##-------------------------------------------------------------------
##-------------------------------------------------------------------
##-------------------------------------------------------------------
def printPairs( pairs, mali, map_new2old, options ):
    """print pairs form codeml."""
    noutput = 0
    for pair in pairs:
        options.stdout.write( "\t".join( map(str, (mali.getEntry(pair.mName2).mId,
                                                   mali.getEntry(pair.mName1).mId,
                                                   pair.mKa, pair.mKs, pair.mKaks,
                                                   pair.mN, pair.mS, "na", "na",
                                                   pair.mKappa, pair.mLogLikelihood,
                                                   pair.mTau ))))
        
        if options.with_rho:
            options.stdout.write( "\t" + "\t".join( map(str, (pair.mRn, pair.mRs, pair.mBranchLength,
                                                              pair.mRn0, pair.mRs0, "na") ) ) )

        if options.with_counts:
            info = Genomics.CalculatePairIndices( mali[pair.mName1], mali[pair.mName2] )
            options.stdout.write( "\t%s" % (str(info)) )

        options.stdout.write( "\t" + pair.mError + "\n" )
        options.stdout.flush()
        noutput += 1
        
    return noutput
##-------------------------------------------------------------------
##-------------------------------------------------------------------
##-------------------------------------------------------------------
def runCodeML( mali, tree, has_non_overlaps, pairs, map_new2old, options ):
    """setup codeml wrapper.

    Sets options and returns a wrapper.
    """

    ids = mali.getIdentifiers()

    ## setup codeml
    codeml_options = {}

    if options.seqtype == "codon":
        codeml_options["seqtype"] = "1"
    elif options.seqtype == "aa":        
        codeml_options["seqtype"] = "2"        
    elif options.seqtype == "trans":
        codeml_options["seqtype"] = "3"

    if options.clean_data:
        codeml_options[ "cleandata" ] = options.clean_data

    if options.omega != None:
        codeml_options[ "omega" ] = str(options.omega)

    if options.kappa != None:
        codeml_options[ "kappa" ] = str(options.kappa)

    if options.fix_kappa:
        codeml_options[ "fix_kappa" ] = "1"

    if options.fix_omega:
        codeml_options[ "fix_omega" ] = "1"

    if options.codon_frequencies != None:
        c = options.codon_frequencies.upper()  
        if c == "UNIFORM":
            a = "0"
        elif c == "F1X4":
            a = "1"
        elif c == "F3X4":
            a = "2"
        elif c == "F61":
            a = "3"
        else:
            a = options.codon_frequencies
        codeml_options[ "CodonFreq" ] = a

    if options.paml_method != None:
        codeml_options[ "paml_method" ] = str(options.method)

    if options.optimization_threshold != None:
        codeml_options[ "Small_Diff" ] = str(options.optimization_threshold)

    ninput, noutput, nskipped = 0, 0, 0
    tstart = time.time()
    
    if pairs and (options.pairwise or has_non_overlaps):
        wrapper = WrapperCodeML.CodeMLPairwise()
        
        ## do pairwise run
        result = WrapperCodeML.CodeMLResultPairs()

        ntotal = (len(ids) * (len(ids) -1 ) ) / 2

        for x,y in pairs:
            m1 = mali.getSequence( ids[x] )                        
            ninput += 1

            temp_mali = Mali.Mali()
            m2 = mali.getSequence( ids[y] )

            temp_mali.addSequence( ids[x], m1.mFrom, m1.mTo, m1.mString )
            temp_mali.addSequence( ids[y], m2.mFrom, m2.mTo, m2.mString )

            ## remove empty columns and masked columns
            if options.clean_mali:
                temp_mali.mGapChars = temp_mali.mGapChars + ("n", "N")
                temp_mali.removeGaps( minimum_gaps = 1, frame=3 )

            if temp_mali.getWidth() < options.min_overlap:
                if options.loglevel >= 1:
                    options.stdlog.write("# pair %s-%s: not computed because only %i residues overlap\n" % (mali.getEntry(ids[x]).mId,
                                                                                                            mali.getEntry(ids[y]).mId,
                                                                                                            temp_mali.getWidth()) )
                nskipped += 1
                continue

            sub_result = wrapper.Run( temp_mali, options = codeml_options, dump=options.dump )
            result.mPairs += sub_result.mPairs

            if options.loglevel >= 1 and ninput % options.report_step == 0:
                options.stdlog.write( "# pairwise computation: %i/%i -> %i%% in %i seconds.\n" % (ninput, ntotal, 100.0 * ninput / ntotal, time.time() - tstart ) )
                options.stdlog.flush()

            noutput += printPairs( sub_result.mPairs,
                                   mali,
                                   map_new2old,
                                   options )

            options.stdout.flush()

        if options.loglevel >= 1:
            options.stdlog.write("# pairwise computation: ninput=%i, noutput=%i, nskipped=%i\n" % (ninput, noutput, nskipped))
            options.stdlog.flush()
                
    else:
        wrapper = WrapperCodeML.CodeML()

        result = wrapper.Run( mali, tree = tree, options = codeml_options, dump=options.dump )

        result_pairs = WrapperCodeML.CodeMLResultPairs()
        result_pairs.fromResult( result )
        noutput += printPairs( result_pairs.mPairs,
                               mali,
                               map_new2old,
                               options )

        l = mali.getLength()
        if options.loglevel >= 1:        
            options.stdlog.write("# input=%i, npairs=%i, noutput=%i\n" % (l, l * (l - 1 ) / 2, len(result_pairs.mPairs) ) )


##-----------------------------------------------------------------------------------
##-----------------------------------------------------------------------------------
##-----------------------------------------------------------------------------------
def checkMatrix( pi, matrix ):
    """check if matrix does fulfill the requirements:
    diagonal sums to -1
    row sum is equal to diagonal entry.
    piQij = pjQji
    """
    codon_table = Bio.Data.CodonTable.standard_dna_table
    codons = codon_table.forward_table.keys()

    ## Matrix needs to be normalized so that branch lenghts in a tree
    ## are effectively measured as expected numbers of nucleotide substituions
    ## per codon
    t = 0.0
    for codon_i in codons:
        t += pi[codon_i] * matrix[codon_i][codon_i]
    # print "check Q diagonal (should be -1): %f" % t        
    assert( "%5.2f" % t == "%5.2f" % -1.0 )
    
    for codon_i in codons:
        t = 0
        for codon_j in codons:
            if codon_i == codon_j: continue
            t += matrix[codon_i][codon_j]
        assert( "%5.2f" % t == "%5.2f" % -matrix[codon_i][codon_i] )
        assert( "%5.2f" % (pi[codon_i] * matrix[codon_i][codon_j]) == "%5.2f" % (pi[codon_j] * matrix[codon_j][codon_i] ) )
        # print "check row %s (should be %f): %f" % (codon_i, -matrix[codon_i][codon_i], t)

##-----------------------------------------------------------------------------------
##-----------------------------------------------------------------------------------
##-----------------------------------------------------------------------------------
def getSequencesFromStk( filename):
    """quick parse through stk file.
    
    Only gets sequences."""
    sequences = {}
    for line in open(filename, "r"):
        if line[0] == '#': continue
        if line[:2] == '//': break
        id, s = re.split('\s+', line[:-1])
        if id not in sequences:
            sequences[id] = []
        sequences[id].append(s)
    
    for id, s in sequences.items():
        sequences[id] = "".join(s)
        
    return sequences

##-----------------------------------------------------------------------------------
##-----------------------------------------------------------------------------------
##-----------------------------------------------------------------------------------
def setFrequencies( model, data):
    """set frequencies in a model according to those observed in data (F3X4).
    """
    sequences = getSequencesFromStk( data )

    frequencies = Codons.getFrequenciesPerCodonPosition( sequences.values() )

    ## build a dummy grammar to insert frequencies
    dummy_grammar = XGram.Model.Grammar()
    for x in range(0,3):
        params = []
        for a in ('a', 'c', 'g', 't'):
            params.append( ("p%s%i" % (a.lower(), x), frequencies[x][a.upper()]) )
        dummy_grammar.addVariable( params )
        
    model.mGrammar.copyParameters( dummy_grammar, 
                                   ignore_missing = True)

##-----------------------------------------------------------------------------------
##-----------------------------------------------------------------------------------
##-----------------------------------------------------------------------------------
def setFixedRates( model, rates):
    """set rates in a model to specified values and fix them."""

    for x in range(len(rates)):
        k = "K%i" % x
        model.mGrammar.setParameter( k, rates[x] )
        model.mGrammar.moveVariableToConst( k )

##-----------------------------------------------------------------------------------
##-----------------------------------------------------------------------------------
##-----------------------------------------------------------------------------------
def writeModel( grammar, section, options):
    """write a model to output file."""
    if section in options.write or "all" in options.write:
        outfile = open( options.output_pattern % section, "w" )
        outfile.write( "%s\n" % grammar.getGrammar())
        outfile.close()

def prepareGrammar( xgram, mali, options ):
    """prepare grammar for custom grammars."""

    ids = mali.getIdentifiers()

    fh, filename = tempfile.mkstemp()
    os.close(fh)
    outfile = open(filename, "w" )
    mali.writeToFile( outfile, format="stockholm",
                      write_ranges = False,
                      options = ("#=GF NH (%s:1.0)%s;" % tuple(ids),) )
    outfile.close()
    
    if options.xrate_model == "sn":
        infile = open(XGram.PATH_DATA + "/sn.eg", "r")
        input_model = XGram.Parser.parseGrammar( infile.readlines() )                

    elif options.xrate_model == "akaksgc":
        infile = open(XGram.PATH_DATA + "/akaksgc.eg", "r")
        input_model = XGram.Parser.parseGrammar( infile.readlines() )

    elif options.xrate_model in ("f3x4-two", "f3x4-four", "f3x4-fourproducts" ):
        input_model = Codons.buildCodonML(codon_model = options.xrate_model,
                                          fix_kappa = options.fix_kappa,
                                          fix_omega = options.fix_omega )

    if options.xrate_model in ("ef3x4-four",):

        sequences = getSequencesFromStk( filename )
        frequencies = Codons.getFrequenciesPerCodonPosition( sequences.values() )

        codon_frequencies = {}
        if options.xrate_insert_frequencies:
            for c1 in ('A', 'C', 'G', 'T'):
                for c2 in ('A', 'C', 'G', 'T'):
                    for c3 in ('A', 'C', 'G', 'T'):
                        codon = "".join( (c1,c2,c3) )
                        if not Genomics.IsStopCodon(codon):
                            codon_frequencies[codon] = frequencies[0][c1] * frequencies[1][c2] * frequencies[2][c3]

            total = sum( codon_frequencies.values())
            for k,v in codon_frequencies.items():
                codon_frequencies[k] /= total
        else:
            for c1 in ('A', 'C', 'G', 'T'):
                for c2 in ('A', 'C', 'G', 'T'):
                    for c3 in ('A', 'C', 'G', 'T'):
                        codon = "".join( (c1,c2,c3) )
                        codon_frequencies[codon] = 1/61.0

        input_model = Codons.buildCodonML(codon_model = "codons-four",
                                          codon_frequencies = codon_frequencies,
                                          fix_kappa = options.fix_kappa,
                                          fix_omega = options.fix_omega )

    else:

        if options.xrate_insert_frequencies:
            setFrequencies( input_model, filename )

        if options.xrate_fix_frequencies:
            for char in ('a', 'c', 'g', 't'):
                for x in (0, 1, 2):
                    param = "p%s%i" % (char, x)
                    input_model.mGrammar.moveVariableToConst( param )

    if options.dump:
        options.stdlog.write( "## input model:\n%s\n" % input_model.getGrammar() )
                    
    writeModel( input_model, "input", options )
    
    t1 = time.time()

    result = xgram.train( input_model, filename )
    
    t2 = time.time()
    
    trained_model = result.getModel()

    if options.dump:
        options.stdlog.write( "## trained model:\n%s\n" % trained_model.getGrammar() )

    writeModel( trained_model, "trained", options )

    return result, mali, ids

def evaluateGrammar( trained_model ):
    
    ## retrieve the terminal frequencies for all codons
    xpi = trained_model.evaluateTerminalFrequencies()[('COD0', 'COD1', 'COD2')]

    pi = {}
    for codon, f in xpi.items():
        pi["".join(codon).upper()] = f

    ## retrieve the rate matrix for all codons
    xmatrix = trained_model.evaluateRateMatrix()[('COD0', 'COD1', 'COD2')]
    matrix = {}
    for codon1, v in matrix.items():
        x = {}
        for codon2, f in matrix[codon1].items():
            x["".join(codon2).upper()] = v
        matrix["".join(codon1).upper()] = x

    return pi, matrix

##-------------------------------------------------------------------
##-------------------------------------------------------------------
##-------------------------------------------------------------------
def runXrateSN( xgram, mali, options ):
    """run xrate using Ians sn.eg grammar."""

    result, mali, ids = prepareGrammar( xgram, mali, options )
    trained_model = result.getModel()
    
    pi, matrix = evaluateGrammar( trained_model )
    
    def getQMatrix( pi, k, s, n):
        """build a q matrix.

        Diagonal elements are set to the negative of the row sums.
        The matrix is normalized such that trace of the matrix is -1.
        """

        codons = Bio.Data.CodonTable.standard_dna_table.forward_table.keys()

        Q = initializeQMatrix( codons )

        trace = 0.0
        for codon_i in codons:
            row_sum = 0.0
            for codon_j in codons:
                if codon_i == codon_j: continue

                is_single, is_synonymous, is_transition = RateEstimation.evaluateCodonPair( codon_i, codon_j )
                
                if not is_single: continue

                if is_synonymous:
                    if is_transition:
                        v = s
                    else:
                        v = s * k
                else:
                    if is_transition:
                        v = n
                    else:
                        v = n * k

                v *= pi[codon_j]
                Q[codon_i][codon_j] = v
                row_sum += v
                
            Q[codon_i][codon_i] = -row_sum
            trace += pi[codon_i] * row_sum

        for codon_i in codons:
            for codon_j in codons:
                Q[codon_i][codon_j] /= trace
                
        return Q, trace

    s = trained_model.mGrammar.getParameter( 's' )
    n = trained_model.mGrammar.getParameter( 'n' )    
    k = trained_model.mGrammar.getParameter( 'k' )    
    not_k = trained_model.mGrammar.getParameter( 'not_k' )    

    Q, t = getQMatrix( pi, k, s, n )
    Q0, t0 = getQMatrix( pi, k, 1, 1 )

    ri, rv, rS, rN = countSubstitutions( pi, Q )
    ri0, rv0, rS0, rN0 = countSubstitutions( pi, Q0 )    

    kappa = ri / rv
    dS = rS / (3 * rS0) * t
    dN = rN / (3 * rN0) * t
    
    if s == None or n == None:
        o_dn, o_ds, o_omega = "na", "na", "na"
        o_rn, o_rn0, o_rs, o_rs0 = "na", "na", "na", "na"
        o_t, o_t0 = "na", "na"
        o_kappa = "na",
        msg = "estimated rate parameters are zero"
    else:
        o_omega = options.value_format % (n / s)
        o_dn = options.value_format % dN
        o_ds = options.value_format % dS
        o_rn = options.value_format % rN
        o_rs = options.value_format % rS
        o_rn0 = options.value_format % rN0
        o_rs0 = options.value_format % rS0
        o_t = options.value_format % t
        o_t0 = options.value_format % t0
        o_kappa = options.value_format % kappa
        msg = "iter=%i s=%6.4f n=%6.4f k=%6.4f ~k=%6.4f" % (result.getNumIterations(), s, n, k, not_k)

    options.stdout.write( "\t".join( map(str, (mali.getEntry(ids[0]).mId,
                                               mali.getEntry(ids[1]).mId,
                                               o_dn, o_ds, o_omega,
                                               "na", "na", "na", "na",
                                               o_kappa, result.getLogLikelihood(),
                                               "na",
                                               o_rn, o_rs, o_t,
                                               o_rn0, o_rs0, o_t0 ))))

    if options.with_counts:
        info = Genomics.CalculatePairIndices( mali[ids[0]], mali[ids[1]] )
        options.stdout.write( "\t%s" % (str(info)) )

    options.stdout.write( "\t%s\n" % msg )

##-------------------------------------------------------------------
##-------------------------------------------------------------------
##-------------------------------------------------------------------
def runXrateAKaKsGc( xgram, mali, options ):
    """run xrate using Ians sn.eg grammar."""

    result, mali, ids = prepareGrammar( xgram, mali, options )
    trained_model = result.getModel()

    rsi = trained_model.mGrammar.getParameter( 'Rsi' )
    rsv = trained_model.mGrammar.getParameter( 'Rsv' )
    rni = trained_model.mGrammar.getParameter( 'Rni' )
    rnv = trained_model.mGrammar.getParameter( 'Rnv' )    
    msg = "iter=%i Rsi=%6.4f Rsv=%6.4f Rni=%6.4f Rnv=%6.4f" % (result.getNumIterations(), rsi, rsv, rni, rnv )
        
    outputXRateResult( mali, result, rsi, rsv, rni, rnv, msg )
        
##-------------------------------------------------------------------
##-------------------------------------------------------------------
##-------------------------------------------------------------------
def runXrateF3X4( xgram, mali, options ):
    """run F3X4 type of models using my grammars.
    """

    result, mali, ids = prepareGrammar( xgram, mali, options )

    trained_model = result.getModel()

    if options.xrate_model == "f3x4-fourproducts":
            i_rs = trained_model.mGrammar.getParameter( 'Rs' )
            i_rn = trained_model.mGrammar.getParameter( 'Rn' )
            i_ri = trained_model.mGrammar.getParameter( 'Ri' )
            i_rv = trained_model.mGrammar.getParameter( 'Rv' )    
            msg = "iter=%i Rs=%6.4f Rn=%6.4f Ri=%6.4f Rv=%6.4f" % (result.getNumIterations(), i_rs, i_rn, i_ri, i_rv )
            rsi = i_rs * i_ri
            rsv = i_rs * i_rv
            rni = i_rn * i_ri
            rnv = i_rn * i_rv
    elif options.xrate_model == "f3x4-two":
            i_kappa = trained_model.mGrammar.getParameter( 'kappa' )
            i_omega = trained_model.mGrammar.getParameter( 'omega' )
            msg = "iter=%i kappa=%6.4f omega=%6.4f" % (result.getNumIterations(), i_kappa, i_omega)
            rsi = i_kappa 
            rsv = 1.0
            rni = i_kappa * i_omega
            rnv = i_omega
    else:
        if options.fix_omega:
            i_kappa = trained_model.mGrammar.getParameter( 'kappa' )
            i_not_kappa = trained_model.mGrammar.getParameter( 'not_kappa' )
            msg = "iter=%i kappa=%6.4f not_kappa=%6.4f" % (result.getNumIterations(), i_kappa, i_not_kappa)
            rsi = i_kappa
            rsv = i_not_kappa
            rni = i_kappa
            rnv = i_not_kappa
        elif options.fix_kappa:
            i_omega = trained_model.mGrammar.getParameter( 'omega' )
            i_not_omega = trained_model.mGrammar.getParameter( 'not_omega' )
            rsi = i_omega
            rsv = i_omega
            rni = i_not_omega        
            rnv = i_not_omega
            msg = "iter=%i omega=%6.4f not_omega=%6.4f" % (result.getNumIterations(), i_omega, i_not_omega)        
        else:
            rsi = trained_model.mGrammar.getParameter( 'Rsi' )
            rsv = trained_model.mGrammar.getParameter( 'Rsv' )
            rni = trained_model.mGrammar.getParameter( 'Rni' )
            rnv = trained_model.mGrammar.getParameter( 'Rnv' )    
            msg = "iter=%i Rsi=%6.4f Rsv=%6.4f Rni=%6.4f Rnv=%6.4f" % (result.getNumIterations(), rsi, rsv, rni, rnv )
        
    outputXRateResult( mali, result, rsi, rsv, rni, rnv, msg )

#################################################################
#################################################################
#################################################################
#################################################################
#################################################################
def outputXRateResult( mali, result, rsi, rsv, rni, rnv, msg ):
    """output the results of running the Xrate four parameter grammar.
    """
    ids = mali.getIdentifiers()
    
    pi, matrix = RateEstimation.getRateMatrix( result.getModel(), 
                                               terminals=('COD0', 'COD1', 'COD2'))

    if rsi == None:
        o_dn, o_ds, o_omega = "na", "na", "na"
        o_rn, o_rn0, o_rs, o_rs0 = "na", "na", "na", "na"
        o_t, o_t0 = "na", "na"
        o_N, o_S = "na", "na"
        o_kappa = "na",
        msg = "estimated rate parameters are zero"
    else:
        Q, t = RateEstimation.getQMatrix( pi,
                                          Rsi=rsi,
                                          Rsv=rsv,
                                          Rni=rni,
                                          Rnv=rnv )

        ## get rate matrix as if omega was set to 1
        Q0, t0 = RateEstimation.getQMatrix( pi,
                                            Rsi = (rsi + rni) / 2.0,
                                            Rsv = (rsv + rnv) / 2.0,
                                            Rni = (rsi + rni) / 2.0,
                                            Rnv = (rsv + rnv) / 2.0 )

        ## get rate matrix as if kappa was set to 1
        Q1, t1 = RateEstimation.getQMatrix( pi,
                                            Rsi = (rsi + rsv) / 2.0,
                                            Rsv = (rsi + rsv) / 2.0,
                                            Rni = (rni + rnv) / 2.0,
                                            Rnv = (rni + rnv) / 2.0 )

        rI, rV, rS, rN = RateEstimation.countSubstitutions( pi, Q )
        rI0, rV0, rS0, rN0 = RateEstimation.countSubstitutions( pi, Q0 )    
        rI1, rV1, rS1, rN1 = RateEstimation.countSubstitutions( pi, Q1 )    

        # 64.0/61.0 results from the fact that xrate does not normalize
        # the terminals 
        dS = rS / (3 * rS0) * t 
        dN = rN / (3 * rN0) * t 
        
        o_omega = options.value_format % (dN / dS)
        o_dn = options.value_format % dN
        o_ds = options.value_format % dS
        o_rn = options.value_format % rN
        o_rs = options.value_format % rS
        o_rn0 = options.value_format % rN0
        o_rs0 = options.value_format % rS0
        o_t = options.value_format % t
        o_t0 = options.value_format % t0
        o_S = options.value_format % (mali.getNumColumns() * rS0)
        o_N = options.value_format % (mali.getNumColumns() * rN0)

        ## kappa is given normalized by sites like omega
        o_kappa = options.value_format % (rI / rI1 * rV1 / rV)

        ## kappa1 is given by the ratio of the rates NOT normalized by the sites.
        msg += " rI/rV=%f rI0/rV0=%f kappa1=%s" % (rI/rV, 
                                                   rI0/rV0,
                                                   options.value_format % ( (rsi + rni) / (rsv + rnv) ) )
    
    options.stdout.write( "\t".join( map(str, (mali.getEntry(ids[0]).mId,
                                               mali.getEntry(ids[1]).mId,
                                               o_dn, o_ds, o_omega,
                                               o_N, o_S, "na", "na",
                                               o_kappa, result.getLogLikelihood(),
                                               "na" ))))

    if options.with_rho:
        options.stdout.write( "\t" + "\t".join( map(str, (o_rn, o_rs, o_t,
                                                          o_rn0, o_rs0, o_t0 ))))

    if options.with_counts:
        info = Genomics.CalculatePairIndices( mali[ids[0]], mali[ids[1]] )
        options.stdout.write( "\t%s" % (str(info)) )

    options.stdout.write( "\t%s\n" % msg )
    options.stdout.flush()

##-------------------------------------------------------------------
##-------------------------------------------------------------------
##-------------------------------------------------------------------
def runXrate( mali, has_non_overlaps, pairs, map_old2new, options ):
    """run xrate on a multiple alignment."""

    ids = mali.getIdentifiers()

    xgram = XGram.XGram()
    if options.xrate_min_increment:
        xgram.setMinIncrement( options.xrate_min_increment )

    ninput, noutput, nskipped = 0, 0, 0

    ## do pairwise run
    for x,y in pairs:
        m1 = mali.getSequence( ids[x] )            
        ninput += 1
        temp_mali = Mali.Mali()
        m2 = mali.getSequence( ids[y] )

        temp_mali.addSequence( m1.mId, m1.mFrom, m1.mTo, m1.mString )
        temp_mali.addSequence( m2.mId, m2.mFrom, m2.mTo, m2.mString )

        ## remove empty columns and masked columns
        if options.clean_mali:
            temp_mali.mGapChars = temp_mali.mGapChars + ("n", "N")
            temp_mali.removeGaps( minimum_gaps = 1, frame=3 )

        if temp_mali.getWidth() < options.min_overlap:
            if options.loglevel >= 1:
                options.stdlog.write("# pair %s-%s: not computed because only %i residues overlap\n" % (mali.getEntry(ids[x]).mId,
                                                                                                        mali.getEntry(ids[y]).mId,
                                                                                                        temp_mali.getWidth()) )

            nskipped += 1
            continue

        if options.xrate_model in ( "sn", ):
            runXrateSN( xgram, temp_mali, options )
        elif options.xrate_model in ( "akaksgc" ):
            runXrateAKaKsGc( xgram, temp_mali, options )            
        else:
            runXrateF3X4( xgram, temp_mali, options )

        if options.loglevel >= 1 and ninput % options.report_step == 0:
            options.stdlog.write( "# pairwise computation: %i/%i -> %i%% in %i seconds.\n" % (ninput, ntotal, 100.0 * ninput / ntotal, time.time() - tstart ) )
            options.stdlog.flush()
            
        noutput += 1

    if options.loglevel >= 1:
        options.stdlog.write("# pairwise computation: ninput=%i, noutput=%i, nskipped=%i\n" % (ninput, noutput, nskipped))
        options.stdlog.flush()
    
##-------------------------------------------------------------------
##-------------------------------------------------------------------
##-------------------------------------------------------------------
def processMali( mali, options ):

    map_new2old = mali.mapIdentifiers()
    ids = mali.getIdentifiers()
        
    invalid_chars = options.gap_chars + options.mask_chars
    
    has_non_overlaps = False

    pairs = []

    if options.iteration == "all-vs-all":
        for x in range(len(ids)):
            for y in range(0,x):
                pairs.append( (x,y) )
    elif options.iteration == "first-vs-all":
        for y in range(1, len(ids)):
            pairs.append( (0,y) )
    elif options.iteration == "pairwise":
        if len(ids) % 2 != 0:
            raise "uneven number of sequences (%i) not compatible with --iteration=pairwise" % len(ids)
        for x in range(0, len(ids), 2):
            pairs.append( (x, x+1) )
    elif options.iteration == "tree":
        pairs = []
    else:
        raise "unknown iteration mode: %s" % (options.iteration)

    if options.remove_stops:
        for id, entry in mali.items():
            s = entry.mString.upper()
            fragments = []
            for x in range(0,len(s), 3):
                codon = s[x:x+3]
                if Genomics.IsStopCodon( codon ):
                    codon = "NNN"
                    
                fragments.append(codon)
                    
            entry.mString = "".join(fragments)

    for x, y in pairs:
        noverlap = 0
        for a, b in zip( mali[ids[x]],mali[ids[y]] ):
            if a not in invalid_chars and b not in invalid_chars:
                noverlap += 1
                if noverlap >= options.min_overlap:
                    break
        else:
            has_non_overlaps = True
            break
        
    if options.tree:
        tree = TreeTools.Newick2Nexus( options.tree ).trees[0]
        map_old2new = IOTools.getInvertedDictionary( map_new2old, make_unique = True )        
        tree.relabel( map_old2new )
    else:
        tree = None

    if options.method == "paml":
        runCodeML( mali, tree, has_non_overlaps, pairs, map_new2old, options )

    elif options.method == "xrate":
        runXrate( mali, has_non_overlaps, pairs, map_new2old, options )


##-------------------------------------------------------------------
##-------------------------------------------------------------------
##-------------------------------------------------------------------

if __name__ == "__main__":
    
    parser = E.OptionParser( version = "%prog version: $Id: mali2kaks.py 2781 2009-09-10 11:33:14Z andreas $" )

    parser.add_option( "--set-omega", dest="omega", type="float",
                      help="initial omega value.")

    parser.add_option( "--set-kappa", dest="kappa", type="float",
                       help="initial kappa value.")

    parser.add_option( "--fix-kappa", dest="fix_kappa", action = "store_true",
                       help="do not estimate kappa.")

    parser.add_option( "--fix-omega", dest="fix_omega", action = "store_true",
                       help="do not estimate omega.")

    parser.add_option( "--set-codon-frequencies", dest="codon_frequencies", type="choice",
                       choices=("uniform", "fequal", "f3x4", "f1x4", "f61" ),
                       help="set codon frequencies.")
    
    parser.add_option( "--set-method", dest="paml_method", type="int",
                      help="set paml optimization method [0|1].")
    
    parser.add_option( "--set-sequence-type", dest="seqtype", type="choice",
                       choices = ("codon", "aa", "trans"),
                       help="sequence type.")

    parser.add_option( "--set-clean-data", dest="clean_data", type="choice",
                       choices = ("0", "1"),
                       help="PAML should cleanup data:  0=only gaps within pair are removed, 1=columns in the mali with gaps are removed.")

    parser.add_option("--dump", dest="dump", action="store_true",
                      help="dump raw output [%default].")
    
    parser.add_option( "--set-optimization-threshold", dest="optimization_threshold", type="string",
                      help="set paml optimization threshold [%default].")
    
    parser.add_option("-i", "--input-format", dest="input_format", type="choice",
                      choices=("plain", "fasta", "clustal", "stockholm", "phylip" ),
                      help="input format of multiple alignment [%default]."  )

    parser.add_option( "--pairwise", dest="pairwise", action="store_true",
                      help="force pairwise comparison [%default].")

    parser.add_option( "--iteration", dest="iteration", type="choice",
                       choices=("all-vs-all", "first-vs-all", "pairwise", "tree" ),
                       help="iteration mode [%default]." )

    parser.add_option( "--no-clean", dest="clean_mali", action="store_false",
                      help="do not clean multiple alignment before submitting to codeml. It might take too long for very large sequences.")

    parser.add_option( "--method", dest="method", type="choice",
                       choices=("paml", "xrate"),
                       help = "choose method for rate computation [%default]" )

    parser.add_option( "--xrate-model", dest="xrate_model", type="choice", 
                      choices=("f3x4-two", "f3x4-four", "sn", "akaksgc", "ef3x4-four", "f3x4-fourproducts" ),
                      help="models to use [%default]."  )

    parser.add_option("-w", "--write", dest="write", type="choice", action="append",
                      choices=("input_fixed", "trained_fixed", "input_variable", "trained_variable", "all" ),
                      help="output sections to write [%default]."  )

    parser.add_option("-o", "--output-pattern", dest="output_pattern", type="string",
                      help="output pattern for output files [%default]." )

    parser.add_option("--xrate-insert-frequencies", dest="xrate_insert_frequencies", action="store_true",
                      help="estimate codon frequencies from input [%default]." )

    parser.add_option("--xrate-uniform-frequencies", dest="xrate_insert_frequencies", action="store_false",
                      help="use uniform codon frequencies [%default]." )

    parser.add_option("--xrate-fix-frequencies", dest="xrate_fix_frequencies", action="store_true",
                      help="set initial frequencies to const [%default]." )

    parser.add_option("--xrate-estimate-frequencies", dest="xrate_fix_frequencies", action="store_false",
                      help="estimate nucleotide frequencies [%default]." )

    parser.add_option("--xrate-fix-rates", dest="fix_rates", type="string",
                      help="""fix rates to specified values. Note that the number of rates has to match the ones
in the model. Provide values in a comma-separated list [%default].""")

    parser.add_option( "--xrate-min-increment", dest="xrate_min_increment", type = float,
                       help="minimum increment to stop iteration in xrate [%default]." )    

    parser.add_option( "--min-overlap", dest="min_overlap", type="int",
                       help="minimum overlap between a sequence pair in residues [%default].")
    
    parser.add_option( "--with-rho", dest="with_rho", action="store_true",
                      help="output rho values (substitution rates per codon). This requires a patched version of PAML [%default]."  )

    parser.add_option( "--with-counts", dest="with_counts", action="store_true",
                      help="output counts of aligned positions, transitions and transversions [%default]."  )

    parser.add_option( "--remove-stops", dest="remove_stops", action="store_true",
                      help="remove stop codons [%default]."  )

    parser.add_option("--replicates", dest="replicates", type="int",
                      help="in benchmarking mode expect ## replicates [%default]." )

    parser.add_option("--tree", dest="tree", type="string",
                      help="use tree for estimation [%default]." )

    parser.set_defaults(
        input_format="fasta",
        omega = None,
        codon_frequencies = None,
        paml_method = None,
        optimization_threshold = None,
        seqtype = "codon",
        dump = False,
        clean_data = False,
        min_overlap = 60,
        gap_chars = "-.",
        mask_chars = "nN",
        pairwise = False,
        kappa = None,
        fix_kappa = False,
        fix_omega = False,
        clean_mali = True,
        method = "paml",
        report_step = 1000,
        loglevel = 1,
        xrate_insert_frequencies = False,
        xrate_fix_frequencies = False,
        write = [],
        output_pattern = "%s.eg",
        value_format = "%6.4f",
        fix_rates = None,
        xrate_from_parameters = False,
        xrate_model = "f3x4-four",
        with_rho = False,
        with_counts = False,
        iteration = "all-vs-all",
        remove_stops = False,
        xrate_min_increment = 0.000001,
        replicates = None,
        tree = None,
        )

    (options, args) = E.Start( parser )

    if options.method == "xrate":
        # imports for xrate computation
        from XGram.Generator.Prebuilt import Codons
        from XGram.Model import Annotation
        import XGram.Run
        import Bio.Data.CodonTable

        # paml like estimation using xrate
        if options.codon_frequencies == "uniform":
            options.xrate_fix_frequencies = True
            options.xrate_insert_frequencies = False
        elif options.codon_frequencies == "f3x4":
            options.xrate_fix_frequencies = True
            options.xrate_insert_frequencies = True
    elif options.method == "paml":
        if not options.codon_frequencies:
            options.codon_frequencies = "F3X4"
    
    if options.fix_rates: options.fix_rates = map( float, options.fix_rates.split(",") )

    if options.pairwise or options.replicates:
        ## read sequences, but not as a multiple alignment. This permits multiple names.
        mali = Mali.SequenceCollection()
    else:
        mali = Mali.Mali()

    mali.readFromFile( sys.stdin, format = options.input_format )

    E.info("read multiple alignment")
    
    if mali.getLength() == 0:
        raise "refusing to process empty alignment."

    ################################################################
    ################################################################
    ################################################################
    ## setup methods
    ################################################################

    options.stdout.write("seq1\tseq2\tdN\tdS\tdNdS\tN\tS\tdN_err\tdS_err\tkappa\tlnL\ttau" )
        
    if options.with_rho:
        options.stdout.write("\trN\trS\tt\trN0\trS0\tt0" )

    if options.with_counts:
        options.stdout.write("\t%s" % Genomics.SequencePairInfo().getHeader() )

    options.stdout.write( "\terror_str\n")
    
    if options.replicates != None:
        ids = mali.getIdentifiers()
        assert( len(ids) % options.replicates == 0 ) 
        s = len(ids) / options.replicates
        for x in range( 0, len(ids), s ):
            m = Mali.Mali()
            for id in ids[x:x+s]:
                m.addEntry( mali.getEntry( id ) )
            processMali( m, options )
    else:
        processMali( mali, options )

    E.Stop()




################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
mali2predictions.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python mali2predictions.py --help

Type::

   python mali2predictions.py --help

for command line help.

Documentation
-------------

Code
----

'''
USAGE="""convert a multiply aligned genomic segment to predictions.

lower case characters are exons, upper-case characters are introns.
Introns of a size smaller than 6 are frameshifts.
"""

import sys
import os
import string
import re
import optparse

import CGAT.Experiment as E
import CGAT.Prediction as Prediction
import CGAT.IndexedFasta as IndexedFasta
import CGAT.Genomics as Genomics
import CGAT.Mali as Mali
import CGAT.IOTools as IOTools

def processCodon( codon ):
    """process a codon.

    A codon is a list of residues belonging to a codon.
    """
    alignment = []

    max_d = 0
    max_x = 0
    for x in range(1, len(codon)):
        d = codon[x][0] - codon[x-1][0]
        if d > max_d:
            max_d = d
            max_x = x

    if max_d > 6:
        # process introns
        frame = max_x % 3
        if frame:
            alignment.append( ["S", 0, frame] )
        alignment.append( ["5", 0, 2] )
        alignment.append( ["I", 0, max_d - 5] )
        alignment.append( ["3", 0, 2] )        
        if frame:
            alignment.append( ["S", 1, 3-frame] )

    elif max_d == 1 and len(codon) == 3:
        alignment.append( ["M", 1, 3] )

    elif len(codon) < 3:
        alignment.append( ["F", 1, len(codon) ] )

    elif len(codon) == 3:
        d = codon[-1][0] - codon[0][0] + 1
        alignment.append( ["F", 1, d] )
    else:
        raise "untreated"        
    
    # print codon, alignment

    return alignment

##------------------------------------------------------------------------
##------------------------------------------------------------------------
##------------------------------------------------------------------------
def getAlignedColumns( mali, exons ):

    # columns that are aligned, maps to the aligned residue in master
    aligned_columns = {}
    # residues in master sequence that are aligned and exonic, maps to the aligned exonic residue
    aligned_exons = {}
    
    if options.master:
        
        try:
            s = mali[options.master]
        except KeyError:
            raise KeyError("master sequence %s not found in mali." % s )

        n = 0
        r = 0
        for x in range(len(s)):
            c = s[x]
            if c in options.gap_chars:
                continue

            r += 1
            aligned_columns[ x ] = r
            
            if c in string.uppercase:
                continue
            
            aligned_exons[ x ] = n
            n += 1
    else:
        raise "please specify a master to define frames."
    
    return aligned_columns, aligned_exons

##------------------------------------------------------------
if __name__ == '__main__':

    parser = E.OptionParser( version = "%prog version: $Id: mali2predictions.py 2781 2009-09-10 11:33:14Z andreas $",
                                    usage = globals()["__doc__"] )

    parser.add_option( "-g", "--genome-file", dest="genome_file", type="string",
                       help="filename with genome."  )

    parser.add_option( "-l", "--filename-locations", dest="filename_locations", type="string",
                       help="filename with locations"  )

    parser.add_option( "-m", "--master", dest="master", type="string",
                       help="the master determines the frame."  )

    parser.set_defaults(
        filename_locations = None,
        gap_chars = "-.",
        master = None
        )
    
    (options, args) = E.Start( parser, add_pipe_options = True )

    if len(args) > 0:
        print USAGE, "no arguments required."
        sys.exit(2)

    mali = Mali.Mali()

    mali.readFromFile( sys.stdin )

    identifiers = mali.getIdentifiers()

    aligned_columns, aligned_exons = getAlignedColumns( mali, options )

    map_id2location = {}

    if options.filename_locations:
        map_id2location = IOTools.ReadMap( open(options.filename_locations, "r") )
        
    options.stdout.write( Prediction.Prediction().getHeader() + "\n" )

    nid = 1
    
    for identifier in identifiers:

        if options.loglevel >= 2:
            options.stdlog.write("# processing %s\n" % (identifier) )

        entry = mali.getEntry( identifier )
        
        sequence = entry.mString
        if sequence[0] not in string.lowercase:
            raise "all sequences should start with an exon."

        was_exon = True
        d = 0
        alignment = []
        carry_over = 0

        last_codon = []
        codon = []
        nchars_in_codon = 0
        n = 0

        last_master_residue = 0
        master_residue = 0
        for column in range(len(sequence)):

            c = sequence[column]
            is_gap = c in options.gap_chars
            is_aligned = column in aligned_columns
            is_exon = column in aligned_exons

            if is_gap:
                continue

            if is_exon:
                master_residue = aligned_exons[column]                
                codon.append( (n, master_residue) )

            n += 1

            # check if we have a complete codon
            if is_exon:
                # A codon is complete, if it ends at frame 2 or
                # it spans more than one codons in the master.
                # Gaps in the master that are a multiple of 3 are ignored
                d = master_residue - last_master_residue - 1
                
                if master_residue % 3 == 2 or (d % 3 != 0 and d > 0):
                    
                    if last_codon:
                        d = codon[0][0] - last_codon[-1][0] - 1
                        if d > 0:
                            # add in-frame introns
                            if d > 10:
                                alignment.append( ["5", 0, 2] )
                                alignment.append( ["I", 0, d - 4] )
                                alignment.append( ["3", 0, 2] )                                
                            else:
                                raise "untreated case"
                        
                    alignment += processCodon( codon )
                    last_codon = codon
                    codon = []
                    
            last_master_residue = master_residue
            
        last = alignment[0]
        new_alignment = []
        for this in alignment[1:]:
            if this[0] == last[0]:
                last[1] += this[1]
                last[2] += this[2]
                continue
            
            new_alignment.append( last )
            last = this
            
        new_alignment.append(last)

        if options.loglevel >= 4:
            options.stdlog.write("# output=%s\n" % (str(new_alignment)))

        assert( new_alignment[-1][2] % 3 == 0)

        lalignment = sum( map( lambda x: x[2], new_alignment ) )
        
        prediction = Prediction.Prediction()
        
        prediction.mQueryToken = identifier
        
        genomic_sequence = re.sub("[%s]" % options.gap_chars, "", mali[identifier])

        prediction.mPredictionId = nid
        nid += 1
        
        if identifier in map_id2location:

            prediction.mSbjctToken, prediction.mSbjctStrand, sfrom, sto = map_id2location[identifier].split(":")[:4]

            prediction.mSbjctGenomeFrom = int(sfrom) + entry.mFrom
            prediction.mSbjctGenomeTo = int(sto)            

        else:
            prediction.mSbjctToken = "unk"
            prediction.mSbjctStrand = "+"
            prediction.mSbjctGenomeFrom = 0

        prediction.mQueryCoverage = 100
        prediction.mPercentIdentity = 100
        prediction.mPercentSimilarity = 100

        prediction.mQueryLength = prediction.mQueryTo
        
        prediction.mSbjctGenomeTo = prediction.mSbjctGenomeFrom + lalignment
            
        prediction.mMapPeptide2Genome = new_alignment
        prediction.mAlignmentString = string.join( map( \
                                      lambda x: string.join(map(str, x), " "),
                                      prediction.mMapPeptide2Genome), " ")

        prediction.mMapPeptide2Translation, prediction.mTranslation = Genomics.Alignment2PeptideAlignment( \
               prediction.mMapPeptide2Genome, 0, 0, genomic_sequence )

        (prediction.mNIntrons, prediction.mNFrameShifts, prediction.mNGaps, prediction.mNSplits, prediction.mNStopCodons, disruptions) = \
                          Genomics.CountGeneFeatures( 0,
                                                      prediction.mMapPeptide2Genome,
                                                      genomic_sequence )
        
        options.stdout.write( str(prediction) + "\n" )
    
    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
Metagenemark.py
=============================================

:Author: Nick Ilott
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

Parses the 'gff' output that is not strictly in gff format
from MetaGeneMark analysis into gff, fasta and amino acid
sequence files

Usage
-----

Example::

   python formatMetagenemark.py --help

Type::

   python formatMetagenemark.py --help

for command line help.

Documentation
-------------

Takes input from stdin and outputs to stdout

Code
----

'''

import os
import sys
import re
import optparse
import itertools
import CGAT.Experiment as E

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = optparse.OptionParser( version = "%prog version: $Id: script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                                    usage = globals()["__doc__"] )

    parser.add_option("-f", "--format", dest="format", type="choice"
                      , choices=("gff", "fasta", "aa"), help="supply help"  )


    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    if options.format == "gff":
        os.system("grep -v '#' | grep -v '^$'") 

    elif options.format == "aa":
        pattern = "Protein"

    elif options.format == "fasta":
        pattern = "DNA"

    # assume that each sequence line does not just contain the 
    # amino acids A, C, T or G - THIS IS NOT OPTIMAL
    # This is a list of the amino acids that are not in either "DNA" or "A", "C", "T", "G"
    # This assumes that each sequence (line) will contain at least one of these
    amino_acids = ["M", "R", "Q", "E", "H", "I", "L", "K", "F", "P", "S", "W", "Y", "V"]
    result = []
    for line in options.stdin.readlines():
        if not line.startswith("##") or line.find("date") != -1 or line.find("gff") != -1 or line.find("source") != -1: continue
        data = line[2:-1]
        if data.startswith("%s" % pattern):
            name = data
            if result:
                options.stdout.write( ">%s\n%s\n" % (prot_name, "".join(result)))
                result = []
        else:
            if pattern == "Protein":
                if  "".join(map(str, [data.find(x) != -1 for x in amino_acids])).find("True") != -1 and data.find("end") == -1 and data.find("%s" % pattern):
                    result.append(data)
                    prot_name = name
            elif pattern == "DNA":
                if  "".join(map(str, [data.find(x) != -1 for x in amino_acids])).find("True") == -1 and data.find("end") == -1 and data.find("%s" % pattern):
                    result.append(data)
                    prot_name = name
    if result:
        options.stdout.write( ">%s\n%s\n" % (prot_name, "".join(result)) )

    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
qkill.py - kill jobs in the queue
==================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

kill jobs from the sun grid engine according to certain criteria.

Usage
-----

Type::

   python <script_name>.py --help

for command line help.

Code
----

""" 

import os
import sys
import re
import optparse
import subprocess
import CGAT.Experiment as E
import xml.etree.ElementTree
import cStringIO as StringIO

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: cgat_script_template.py 2781 2009-09-10 11:33:14Z andreas $", usage = globals()["__doc__"] )

    parser.add_option("-p", "--pattern", dest="pattern", type="string",
                      help="jobs matching `pattern` in their job description will be killed [default=%default]."  )

    parser.add_option("-n", "--dry-run", dest="dry_run", action="store_true",
                      help="do dry run, do not kill [default=%default]."  )

    parser.set_defaults(
        pattern = None,
        dry_run = False,
        )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    output = StringIO.StringIO(subprocess.Popen(["qstat", "-xml"], stdout=subprocess.PIPE).communicate()[0])
    
    tree = xml.etree.ElementTree.ElementTree( file = output )

    ntested = 0
    to_kill = set()

    if options.pattern:
        pattern = re.compile(options.pattern)
    else:
        pattern = None
    
    for x in tree.getiterator( "job_list"):
        ntested += 1
        id = x.find( "JB_job_number" ).text
        name = x.find( "JB_name" ).text
        if pattern and pattern.search( name ):
            to_kill.add( id )
            
    nkilled = len(to_kill)
    if not options.dry_run:
        p = subprocess.Popen(["qdel", ",".join( to_kill)], stdout=subprocess.PIPE)
        stdout, stderr = p.communicate()
    
    E.info( "ntested=%i, nkilled=%i" % (ntested, nkilled) )

    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

Convert BAM files into BED files.

Usage
-----

Example::

   python bam2bed.py in.bam > out.bed

This command converts the BAM file in.bam into a BED file named out.bed.

Type::

   python bam2bed.py --help

for command line help.

Documentation
-------------

This tool converts BAM files into BED files supplying the intervals for each read in the BAM file.  BAM files must
have a corresponding index file ie. example.bam and example.bam.bai

For example::

   samtools view example.bam

   READ1     163     1       13040   15      76M     =       13183   219     ...
   READ2     99      1       13120   0       76M     =       13207   163     ...
   READ1     83      1       13183   7       76M     =       13040   -219    ...
   READ2     147     1       13207   0       76M     =       13120   -163    ...

   python bam2bed.py example.bam 

   1       13039   13115   READ1     15      +
   1       13119   13195   READ2     0       +
   1       13182   13258   READ1     7       -
   1       13206   13282   READ2     0       -

Options
^^^^^^^
+-------------------+------------------------------------------------------------------+
|--region, -r       |output read intervals that overlap a specified region             |
+-------------------+------------------------------------------------------------------+
|--merge-pairs, -m  |merge paired-end reads and output interval for entire fragment    |
+-------------------+------------------------------------------------------------------+
|--max-insert-size  |only merge if insert size is less than specified no. of bases     |
+-------------------+------------------------------------------------------------------+
|--min-insert-size  |only merge if insert size is greater than specified no. of bases  |
+-------------------+------------------------------------------------------------------+

For example,

To output read intervals that overlap chromosome 1, coordinates 13000-13100::

   samtools view example.bam

   READ1     163     1       13040   15      76M     =       13183   219     ...
   READ2     99      1       13120   0       76M     =       13207   163     ...
   READ1     83      1       13183   7       76M     =       13040   -219    ...
   READ2     147     1       13207   0       76M     =       13120   -163    ...

   python bam2bed.py example.bam --region '1:13000:13100'

   1       13039   13115   READ1     15      +

To merge paired-end reads and output fragment interval ie. leftmost mapped base to rightmost mapped base::

   python bam2bed.py example.bam --merge-pairs

   1       13119   13282   READ2     0       +
   1       13039   13258   READ1     7       +
   
Code
----

""" 

import os
import sys
import re
import optparse
import pysam

import CGAT.Experiment as E

try:
    import pyximport
    pyximport.install(build_in_temp=False)
    import _bam2bed
except ImportError:
    import CGAT._bam2bed as _bam2bed

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id$", usage = globals()["__doc__"] )


    parser.add_option("-r", "--region", dest="region", type="string",
                      help = "samtools region string [default=%default]. " )

    parser.add_option("-m", "--merge-pairs", dest="merge_pairs", action="store_true",
                      help = "merge paired-ended reads into a single bed interval [default=%default]. " )

    parser.add_option( "--max-insert-size", dest="max_insert_size", type = "int",
                      help = "only merge paired-end reads if they are less than # bases apart. " 
                             " 0 turns off this filter [default=%default]. " )

    parser.add_option( "--min-insert-size", dest="min_insert_size", type = "int",
                       help = "only merge paired-end reads if they are at least # bases apart. "
                              " 0 turns off this filter [default=%default]. " )

    parser.add_option( "--bed-format", dest="bed_format", type = "choice", 
                       choices = ('3','4','5','6'),
                       help = "bed format to output. "
                              " [default=%default]" )

    parser.set_defaults(
        region = None,
        call_peaks = None,
        merge_pairs = None,
        min_insert_size = 0,
        max_insert_size = 0,
        bed_format = '6',
        )

    (options, args) = E.Start( parser, argv = argv )

    if len(args) != 1:
        raise ValueError( "no samfile specified - see --help for usage" )
    
    samfile = pysam.Samfile( args[0], "rb" )

    options.bed_format = int(options.bed_format)

    if options.merge_pairs != None:
        counter = _bam2bed.merge_pairs( samfile, 
                                        options.stdout,
                                        min_insert_size = options.min_insert_size,
                                        max_insert_size = options.max_insert_size,
                                        bed_format = options.bed_format )

        options.stdlog.write( "category\tcounts\n%s\n" % counter.asTable() )

    else:
        if options.region != None:
            if args[0] == "-":
                raise ValueError("can't use region with a file from stdin" )
            it = samfile.fetch( region = options.region )
        else:
            # use until_eof. Files from stdin have no index
            it = samfile.fetch( until_eof = True )

        # more comfortable cigar parsing will
        # come with the next pysam release
        BAM_CMATCH = 0
        BAM_CDEL = 2
        BAM_CREF_SKIP = 3
        take = (BAM_CMATCH, BAM_CDEL, BAM_CREF_SKIP)
        outfile = options.stdout

        for read in it:
            if read.is_unmapped: continue

            t = 0
            for op, l in read.cigar:
                if op in take: t += l

            start = read.pos
            if read.is_reverse: strand = "-"
            else: strand = "+"
            # IMS: converted rname to reference name
            outfile.write("%s\t%d\t%d\t%s\t%d\t%c\n" %\
                          ( samfile.getrname(read.rname),
                            read.pos,
                            read.pos+t,
                            read.qname,
                            read.mapq,
                            strand) )            

    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )




################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
Expression.py - wrap various differential expression tools
===========================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This module provides tools for differential expression analysis 
for a variety of methods.

Methods implemented are:

   DESeq
   EdgeR
   cuffdiff

There is also a command line interface. Note that the module is incomplete
as a stand-alone script as it requires to be executed in the context of
an existing pipeline for parameterization.

Usage
-----

Documentation
-------------

Code
----

'''

import math
import numpy
import sys
import os
import collections
import itertools

from rpy2.robjects import r as R
import rpy2.robjects as ro
import rpy2.robjects.numpy2ri

try:
    import CGAT.Experiment as E
    import CGAT.Pipeline as P
    import CGAT.Database as Database
    import CGAT.IOTools as IOTools
    import CGAT.Stats as Stats
except ImportError:
    import Experiment as E
    import Pipeline as P
    import Database,IOTools,Stats

import sqlite3

try:
    PARAMS = P.getParameters()
except IOError:
    pass


def buildProbeset2Gene( infile, 
                        outfile, 
                        database = "hgu133plus2.db",
                        mapping = "hgu133plus2ENSEMBL" ):
    
    '''build map relating a probeset to an ENSEMBL gene_id'''
    
    R.library( database )

    # map is a Bimap object
    m = R(mapping)
    
    result = R.toTable(m)

    outf = open( outfile, "w")
    outf.write( "probe_id\tgene_id\n" )
    for probeset_id, gene_id in zip(result["probe_id"], 
                                    result["ensembl_id"] ):
        outf.write( "%s\t%s\n" % (probeset_id, gene_id))
    outf.close()

    E.info( "written %i mappings to %s: probes=%i, genes=%i" % \
                (len(result),
                 outfile,
                 len(set(result["probe_id"])),
                 len(set(result["ensembl_id"])) ) )

GeneExpressionResult = collections.namedtuple( "GeneExpressionResult", \
                                                   "test_id treatment_name treatment_mean treatment_std " \
                                                   " control_name control_mean control_std " \
                                                   " pvalue qvalue l2fold fold significant status" )


def writeExpressionResults( outfile, result ):
    '''output expression results table.'''
    outfile.write( "%s\n" % "\t".join(GeneExpressionResult._fields))
    for x in result:
        outfile.write("%s\n" % "\t".join( map(str,x)))

class WelchsTTest(object):
    '''base class for computing expression differences.
    '''

    def __call__(self, 
                 probesets, 
                 treatments, 
                 controls):
        
        assert len(probesets) == len(treatments[0])
        assert len(probesets) == len(controls[0])
        
        nskipped = 0
        results = []

        for probeset, treatment, control in zip( probesets, zip(*treatments), zip(*controls) ):

            nval1, nval2 = len(treatment), len(control)
            mean1, mean2 = numpy.mean(treatment), numpy.mean(control)
            stddev1, stddev2 = numpy.std(treatment), numpy.std(control)
            
            try:
                s = Stats.doWelchsTTest( nval1, mean1, stddev1,
                                         nval2, mean2, stddev2,
                                         alpha = 0.05 )
            except ValueError:
                E.warn("expressionDifferences: standard deviations are 0 for probeset %s - skipped" % probeset )
                nskipped += 1
                continue

            s.mProbeset = probeset
            results.append( s )

        qvalues =  Stats.doFDR( [x.mPValue for x in results ] ).mQValues

        for s, qvalue in zip(results, qvalues ):
            s.mQValue = qvalue

        return results, nskipped

class SAMR( object ):
    '''SAM analysis of microarray data.

    Use the Two-Class Unpaired Case Assuming Unequal Variances.

    This uses the samr library.

    Significant genes are either called at *fdr* or the
    top *ngenes* are returned.

    *treatments* and *control* are arrays of
    arrays of expression values.

    See 

    https://stat.ethz.ch/pipermail/bioconductor/2008-July/023251.html

    for an explanation of the differences between siggens SAM
    and Excel SAM. This version is parameterised to reproduce Excel SAM
    by setting::
       
       var.equal = TRUE
       med = TRUE

    .. note:: 
        SAM requires log2 scaled expression levels.
    '''
    
    def __call__(self, probesets, 
                 treatments, 
                 controls,
                 pattern = None,
                 fdr = 0.10,
                 ngenes = None,
                 npermutations = 1000,
                 ndelta=10,
                 method = "ttest" ):

        if ngenes and fdr:
            raise ValueError( "either supply ngenes or fdr, but not both.")

        R.library("samr")

        m = numpy.matrix( treatments + controls )
        m = numpy.transpose(m)
        labels = numpy.array([1] * len(treatments) + [2] * len(controls))

        R.assign("x", numpy.array(m))
        R.assign("y", labels)
        R.assign("probesets", probesets)

        data = R('''data=list( x=x, y=y, geneid=1:length(probesets), genenames=probesets, logged2=TRUE)''' )
        result = R('''samr.obj<-samr(data,  resp.type="Two class unpaired", nperms=100)''')
        R('''plot(samr.obj, delta=.4)''')

class SAM( object ):
    '''SAM analysis of microarray data.

    Use the Two-Class Unpaired Case Assuming Unequal Variances.

    This uses the siggenes library. Note that there is also
    an rsam package at:

    http://rss.acs.unt.edu/Rdoc/library/samr/html/samr.html

    Significant genes are either called at *fdr* or the
    top *ngenes* are returned.

    *treatments* and *control* are arrays of
    arrays of expression values.

    See 

    https://stat.ethz.ch/pipermail/bioconductor/2008-July/023251.html

    for an explanation of the differences between siggens SAM
    and Excel SAM. To parameterize the FDR to excel sam, set the
    flag *use_excel_sam*.

    .. note:: 
        SAM requires log2 scaled expression levels.

    I ran into trouble using this library. I was not able to
    reproduce the same results from the original SAM study getting
    differences in d and in the fdr.

    fold change is treatment / control.

    '''
    
    def __call__(self, probesets, 
                 treatments, 
                 controls,
                 pattern = None,
                 fdr = 0.10,
                 ngenes = None,
                 npermutations = 1000,
                 ndelta=10,
                 method = "ttest",
                 use_excel_sam = False,
                 treatment_label = "treatment",
                 control_label = "control" ):
        
        if ngenes and fdr:
            raise ValueError( "either supply ngenes or fdr, but not both.")
        
        R.library("siggenes")

        m = numpy.matrix( treatments + controls )
        m = numpy.transpose(m)

        E.debug( "build expression matrix: %i x %i" % m.shape )

        labels = numpy.array([1] * len(treatments) + [0] * len(controls))
        ## 1000 permutations for P-Values of down to 0.0001. Setting this
        ## to a high value improved reproducibility of results.

        kwargs = {}
        # kwargs set to replicate excel SAM
        if use_excel_sam:
            kwargs.update( { "control" : R('''samControl( lambda = 0.5, n.delta = %(ndelta)s ) ''' % locals()),
                             "med" : True,
                             "var.equal": True } )
        else:
            kwargs.update( { "control" : R('''samControl( n.delta = %(ndelta)s ) ''' % locals()) }, 
                           )

        # the option B needs to be not set if wilc.stat is chosen

        if method == "ttest":
            kwargs["method"] = R('''d.stat''')
            kwargs["B"] = npermutations            
        elif method == "wilc":
            kwargs["method"] = R('''wilc.stat''')
        elif metod == "cat":
            kwargs["method"] = R('''cat.stat''')
        else:
            raise ValueError("unknown statistic `%s`" % method )

        E.info( "running sam with the following options: %s" % str(kwargs) )
        
        a = R.sam( numpy.array(m),
                   labels,
                   gene_names=numpy.array(probesets),
                   **kwargs )
        
        # E.debug("%s" % str(a))

        R.assign( "a", a )

        fdr_data = collections.namedtuple( "sam_fdr", ("delta", "p0", "false", "significant", "fdr", "cutlow","cutup", "j2","j1" ) )
        cutoff_data = collections.namedtuple( "sam_cutoff", ("delta", "significant", "fdr"))
        gene_data = collections.namedtuple( "sam_fdr", ("row","dvalue","stddev","rawp","qvalue","rfold" ) )

        def _totable( robj ):
            '''convert robj to a row-wise table.'''
            s = numpy.matrix( robj )
            t = [ numpy.array(x).reshape(-1,) for x in s ]
            return t

        # extract the fdr values
        # returns R matrix
        t = _totable( a.do_slot('mat.fdr') )
        assert len(t[0]) == len(fdr_data._fields)
        # for x in t: E.debug( "x=%s" % str(x))
        fdr_values = [ fdr_data( *x ) for x in t ]

        # find d cutoff
        if fdr != None and fdr > 0:
            s = numpy.matrix( R.findDelta( a, fdr ) )
            try:
                cutoffs = [ cutoff_data( *numpy.array(x).reshape(-1,) ) for x in s ]
                E.debug( "sam cutoffs for fdr %f: %s" % (fdr, str(cutoffs) ) )
                cutoff = cutoffs[-1]
            except TypeError:
                E.debug( "could not get cutoff" )
                cutoff = None
        elif ngenes:
            s = numpy.matrix( R.findDelta( a, ngenes ) )
            try:
                cutoffs = [ cutoff_data( *numpy.array(x).reshape(-1,) ) for x in s ]
                E.debug( "sam cutoffs for fdr %f: %s" % (fdr, str(cutoffs) ) )
                cutoff = cutoffs[-1]
            except TypeError:
                E.debug( "could not get cutoff" )
                cutoff = None
        else:
            raise ValueError("either supply ngenes or fdr")

        # collect (unadjusted) p-values and qvalues for all probesets
        pvalues = dict( zip( probesets, R('''a@p.value''') ) )
        qvalues = dict( zip( probesets, R('''a@q.value''') ) )

        if pattern:
            outfile = pattern % "sam.pdf"
            R.pdf(outfile)
            if cutoff:
                R.plot( a, cutoff.delta )
            else:
                R.plot( a )
            R['dev.off']()
        
        siggenes = {}        
        significant_genes = set()
        if cutoff != None:
            E.debug( "using cutoff %s" % str(cutoff) )
            
            summary = R('''summary( a, %f )''' % cutoff.delta)

            # summary = R.summary( a, cutoff.delta )
            R.assign( "summary", summary )

            significant_genes = set( [probesets[int(x)-1] for x in R('''summary@row.sig.genes''')] )
            # E.debug( "significant genes=%s" % str(significant_genes))
            
            r_result = zip(*_totable( summary.do_slot( 'mat.sig' ) ))
            
            if len(r_result) > 0:

                assert len(r_result[0]) == 6, "expected six columns from siggenes module, got: %s" % len(r_result[0])
                
                for x in r_result:
                    if x[4] > fdr:
                        E.warn( "%s has qvalue (%f) larger than cutoff, but is significant significant." % (str(x), x[4]))
                            
                # except TypeError:
                #     # only a single value
                #     x = [r_result[y] for y in ("Row", "d.value", "stdev", "rawp", "q.value", "R.fold") ]
                #     if x[4] > fdr:
                #         E.warn( "%s has qvalue (%f) larger than cutoff, but is called significant." % (str(x), x[4]))

                siggenes[probesets[int(x[0])-1]] = gene_data( *x )                

        else:
            E.debug( "no cutoff found - no significant genes." )
            
        genes = []
        for probeset, treatment, control in zip( probesets, zip(*treatments), zip(*controls) ):

            mean1, mean2 = numpy.mean(treatment), numpy.mean(control)

            if probeset in siggenes:
                s = siggenes[probeset]
                pvalue = s.rawp
                qvalue = s.qvalue
            else:
                pvalue = pvalues[probeset]
                qvalue = qvalues[probeset]

            significant = (0,1)[probeset in significant_genes]

            genes.append( GeneExpressionResult._make( (probeset,
                                                       treatment_label,
                                                       mean1,
                                                       numpy.std( treatment ),
                                                       control_label,
                                                       mean2,
                                                       numpy.std( control ),
                                                       pvalue,
                                                       qvalue,
                                                       mean1 - mean2,
                                                       math.pow(2,mean1 - mean2),
                                                       significant,
                                                       "OK" ) ) )

        return genes, cutoff, fdr_values



#########################################################################
#########################################################################
#########################################################################
def loadTagData( tags_filename, design_filename ):
    '''load tag data for deseq/edger analysis.
    
    *Infile* is a tab-separated file with counts.

    *design_file* is a tab-separated file with the
    experimental design with four columns::

      track   include group   pair
      CW-CD14-R1      0       CD14    1
      CW-CD14-R2      0       CD14    1
      CW-CD14-R3      1       CD14    1
      CW-CD4-R1       1       CD4     1
      FM-CD14-R1      1       CD14    2
      FM-CD4-R2       0       CD4     2
      FM-CD4-R3       0       CD4     2
      FM-CD4-R4       0       CD4     2

    track
        name of track - should correspond to column header in *infile*
    include
        flag to indicate whether or not to include this data
    group
        group indicator - experimental group
    pair
        pair that sample belongs to (for paired tests)

    This method creates various R objects:

    countsTable : data frame with counts.
    groups : vector with groups
    pairs  : vector with pairs

    '''

    E.info( "loading tag data from %s" % tags_filename)

    R( '''counts_table = read.delim( '%(tags_filename)s', 
                                     header = TRUE,
                                     row.names = 1,
                                     stringsAsFactors = TRUE,
                                     comment.char = '#' )''' % locals() )

    E.info( "read data: %i observations for %i samples" % tuple(R('''dim(counts_table)''')))
    E.debug( "sample names: %s" % R('''colnames(counts_table)'''))

    # Load comparisons from file
    R('''pheno = read.delim( '%(design_filename)s', 
                             header = TRUE, 
                             stringsAsFactors = TRUE,
                             comment.char = '#')''' % locals() )

    # Make sample names R-like - substitute - for .
    R('''pheno[,1] = gsub('-', '.', pheno[,1]) ''')
    E.debug( "design names: %s" % R('''pheno[,1]'''))
    
    # Ensure pheno rows match count columns
    pheno = R('''pheno2 = pheno[match(colnames(counts_table),pheno[,1]),,drop=FALSE]''' )
    missing = R('''colnames(counts_table)[is.na(pheno2)][1]''')
    if missing:
        E.warn( "missing samples from design file are ignored: %s" % missing)
        
    # Subset data & set conditions
    R('''includedSamples <- !(is.na(pheno2$include) | pheno2$include == '0') ''')
    E.debug( "included samples: %s" % R('''colnames(counts_table)[includedSamples]''') )
    R('''countsTable <- counts_table[ , includedSamples ]''')
    R('''groups <- factor(pheno2$group[ includedSamples ])''')
    R('''conds <- pheno2$group[ includedSamples ]''')
    R('''pairs = factor(pheno2$pair[ includedSamples ])''')
 
    E.info( "filtered data: %i observations for %i samples" % tuple( R('''dim(countsTable)''') ) )

def filterTagData( min_sample_counts = 10):
    '''filter tag data.'''
    
    # Remove windows with no data
    R( '''max_counts = apply(countsTable,1,max)''' )
    R( '''countsTable = countsTable[max_counts>0,]''')
    E.info( "removed %i empty rows" % tuple( R('''sum(max_counts == 0)''') ) )
    observations, samples = tuple( R('''dim(countsTable)'''))
    E.info( "trimmed data: %i observations for %i samples" % (observations, samples ))

    # remove samples without data
    R('''max_counts = apply(countsTable,2,max)''' )
    empty_samples = tuple(R('''max_counts < %i''' % min_sample_counts))
    sample_names = R('''colnames(countsTable)''')
    nempty_samples = sum( empty_samples)
    if nempty_samples:
        E.warn( "%i empty samples are being removed: %s" % \
                    (nempty_samples, ",".join( [sample_names[x] for x,y in enumerate( empty_samples) if y]) ) )
        R('''countsTable <- countsTable[, max_counts >= %i]''' % min_sample_counts)
        R('''groups <- groups[max_counts >= %i]''' % min_sample_counts)
        R('''pairs <- pairs[max_counts >= %i]''' % min_sample_counts)
        observations, samples = tuple( R('''dim(countsTable)'''))

    return observations, samples

def groupTagData(ref_group = None):
    '''compute groups and pairs from tag data table.'''

    # Relevel the groups so that the reference comes first
    if not ref_group is None:
        R('''groups <- relevel(groups, ref = "%s")''' % ref_group)

    groups = R('''levels(groups)''')
    pairs = R('''levels(pairs)''')
    
    # Test if replicates exist - at least to samples pre replicate
    min_per_group = R('''min(table(groups)) ''')[0]
    has_replicates = min_per_group >= 2

    # Test if pairs exist:
    npairs = R('''length(table(pairs)) ''')[0]
    has_pairs = npairs == 2

    # at least two samples per pair
    if has_pairs:
        min_per_pair = R('''min(table(pairs)) ''')[0]
        has_pairs = min_per_pair >= 2

    return groups, pairs, has_replicates, has_pairs
    
def plotHeatmap():
    '''plot a heatmap.'''
    
    R('''dists <- dist( t(as.matrix(countsTable)) )''')
    R('''heatmap( as.matrix( dists ), symm=TRUE )''' )

def runEdgeR( infile, 
              design_file, 
              outfile, 
              outfile_prefix = "edger.",
              fdr = 0.1,
              prefix = "",
              dispersion = None,
              ref_group = None
              ):
    '''run DESeq on.

    See loadTagData on the input form format for *infile* and
    *design_file*.

    Results are stored in *outfile* and files prefixed by *outfile_prefix*.

    The dispersion is usually measuered from replicates. If there are no 
    replicates, you need to set the *dispersion* explicitely.

    See page 13 of the EdgeR user guide::

       2. Simply pick a reasonable dispersion value, based on your experience with similar data,
       and use that. Although subjective, this is still more defensible than assuming Poisson
       variation. Typical values are dispersion=0.4 for human data, dispersion=0.1 for data
       on genetically identical model organisms or dispersion=0.01 for technical replicates.

    '''
    
    # load library 
    R('''suppressMessages(library('edgeR'))''')

    to_cluster = True

    loadTagData( infile, design_file )

    nobservations, nsamples = filterTagData()

    if nobservations == 0:
        E.warn( "no observations - no output" )
        return

    if nsamples == 0:
        E.warn( "no samples remain after filtering - no output" )
        return

    groups, pairs, has_replicates, has_pairs = groupTagData(ref_group)

    sample_names = R('''colnames(countsTable)''')
    E.info( "%i samples to test at %i observations: %s" % ( nsamples, nobservations,
                                                            ",".join( sample_names)))

    # output heatmap plot
    R.png( '%(outfile_prefix)sheatmap.png' % locals() )
    plotHeatmap()
    R['dev.off']()    

    E.info('running EdgeR: groups=%s, pairs=%s, replicates=%s, pairs=%s' % \
               (groups, pairs, has_replicates, has_pairs))
    
    if has_pairs:
        # output difference between groups
        R.png( '''%(outfile_prefix)sbalance_groups.png''' % locals() )
        first = True
        for g1, g2 in itertools.combinations(groups, 2 ):
            R('''a = rowSums( countsTable[groups == '%s'] ) ''' % g1 )
            R('''b = rowSums( countsTable[groups == '%s'] ) ''' % g2 )
            if first:
                R('''plot( cumsum( sort(a - b) ), type = 'l') ''' )
                first = False
            else:
                R('''lines( cumsum( sort(a - b) )) ''' )

        R['dev.off']()

        R('''library('ggplot2')''')
        R('''library('reshape')''')

        # output difference between pairs within groups
        first = True
        legend = []
        for pair in pairs:
            for g1, g2 in itertools.combinations(groups, 2 ):
                key = "pair_%s_%s_vs_%s" % (pair, g1,g2)
                legend.append( key )
                print R('''colnames( countsTable) ''')
                print R(''' pairs=='%s' ''' % pair)
                print R(''' groups=='%s' ''' % g1)
                R('''a = rowSums( countsTable[pairs == '%s' & groups == '%s'] ) ''' % (pair,g1) )
                R('''b = rowSums( countsTable[pairs == '%s' & groups == '%s'] ) ''' % (pair,g2) )
                R('''c = cumsum( sort(a - b) )''' )
                R('''c = c - min(c)''')
                if first:
                    data = R( '''d = data.frame( %s = c)''' % key)
                    first = False
                else:
                    R('''d$%s = c''' % key)

        # remove row names (gene idenitifiers)
        R('''row.names(d) = NULL''')
        # add numbers of genes (x-axis)
        R('''d$genes=1:nrow(d)''')

        # merge data for ggplot
        R('''d = melt( d, 'genes', variable_name = 'comparison' )''')

        # plot
        R('''gp = ggplot(d)''')
        R('''pp = gp + \
            geom_line(aes(x=genes,y=value,group=comparison,color=comparison))''')
                    
        R.ggsave( '''%(outfile_prefix)sbalance_pairs.png''' % locals() )
        R['dev.off']()

    # build DGEList object
    R( '''countsTable = DGEList( countsTable, group = groups )''' )

    # Relevel groups to make the results predictable - IMS
    if not ref_group is None:
        R('''countsTable$samples$group <- relevel(countsTable$samples$group, ref = "%s")''' % ref_group)
    else:
        #if no ref_group provided use first group in groups
        R('''countsTable$sample$group <- relevel(countsTable$samples$group, ref = "%s")''' % groups[0])

    # calculate normalisation factors
    E.info( "calculating normalization factors" )
    R('''countsTable = calcNormFactors( countsTable )''' )
    E.info( "output")

    # Remove windows with few counts
    # R( '''countsTable = countsTable[rowSums( 
    #          1e+06 * countsTable$counts / 
    #           expandAsMatrix ( countsTable$samples$lib.size, dim(countsTable)) > 1 ) >= 2, ]''')

    E.info( "trimmed data: %i observations for %i samples" % tuple( R('''dim(countsTable)''') ) )

    # output MDS plot
    R.png( '''%(outfile_prefix)smds.png''' % locals() )
    try:
        R('''plotMDS( countsTable )''')
    except rpy2.rinterface.RRuntimeError, msg:
        E.warn( "can not plot mds: %s" % msg)
    R['dev.off']()

    # build design matrix
    if has_pairs:
        R('''design = model.matrix( ~pairs + countsTable$samples$group )''' )
    else:
        R('''design = model.matrix( ~countsTable$samples$group )''' )

    # R('''rownames(design) = rownames( countsTable$samples )''')
    # R('''colnames(design)[length(colnames(design))] = "CD4" ''' )
    
    # fitting model to each tag
    if has_replicates:
        # estimate common dispersion
        R('''countsTable = estimateGLMCommonDisp( countsTable, design )''')
        # estimate tagwise dispersion
        R('''countsTable = estimateGLMTagwiseDisp( countsTable, design )''')
        # fitting model to each tag
        R('''fit = glmFit( countsTable, design )''')
    else:
        # fitting model to each tag
        if dispersion == None:
            raise ValueError( "no replicates and no dispersion" )
        E.warn("no replicates - using a fixed dispersion value" )
        R('''fit = glmFit( countsTable, design, dispersion = %f )''' % dispersion )

    # perform LR test
    R('''lrt = glmLRT(fit)''' )

    E.info("Generating output")

    
    # output cpm table
    R('''library(reshape2)''')
    R('''countsTable.cpm <- cpm(countsTable,  normalized.lib.sizes=TRUE)''')
    R('''countsTable.cpm.melt <- melt(countsTable.cpm)''')
    R('''names(countsTable.cpm.melt) <- c("id","sample","ncpm")''')
    R('''write.table(countsTable.cpm.melt, file="%(outfile_prefix)scpm.tsv", sep = "\t", row.names=FALSE, quote=FALSE)''' % locals())

    # compute adjusted P-Values
    R('''padj = p.adjust( lrt$table$PValue, 'BH' )''' )

    isna = R["is.na"]

    rtype = collections.namedtuple( "rtype", "lfold logCPM LR pvalue" )

    # output differences between pairs
    R.png( '''%(outfile_prefix)smaplot.png''' % locals() )
    R('''plotSmear( countsTable, pair=c('%s') )''' % "','".join( groups) )
    R('''abline( h = c(-2,2), col = 'dodgerblue') ''' )
    R['dev.off']()

    # I am assuming that logFC is the base 2 logarithm foldchange.
    # Parse results and parse to file
    results = []
    counts = E.Counter()
    
    for interval, data, padj in zip( R('''rownames(lrt$table)'''),
                                     zip( *R('''lrt$table''')), 
                                     R('''padj''')) :
        d = rtype._make( data )

        counts.input += 1

        # set significant flag
        if padj <= fdr: 
            signif = 1
            counts.significant += 1
            if d.lfold > 0:
                counts.significant_over += 1
            else:
                counts.significant_under += 1
        else: 
            signif = 0
            counts.insignificant += 1

        if d.lfold > 0:
            counts.all_over += 1
        else:
            counts.all_under += 1
        
        if isna( d.pvalue ): status = "OK"
        else: status = "FAIL"

        counts[status] += 1

        try:
            fold = math.pow( 2.0, d.lfold )
        except OverflowError:
            E.warn( "%s: fold change out of range: lfold=%f" % (interval, d.lfold ))
            # if out of range set to 0
            fold = 0
            
        # fold change is determined by the alphabetical order of the factors. 
        # Is the following correct?
        results.append( GeneExpressionResult._make( ( \
                    interval,
                    groups[1],
                    d.logCPM,
                    0,
                    groups[0],
                    d.logCPM,
                    0,
                    d.pvalue,
                    padj,
                    d.lfold,
                    fold,
                    str(signif),
                    status) ) )
            
    if outfile == sys.stdout:
        writeExpressionResults( outfile, results )
    else:
        with IOTools.openFile( outfile, "w" ) as outf:
            writeExpressionResults( outf, results )

    outf = IOTools.openFile( "%(outfile_prefix)ssummary.tsv" % locals(), "w" )
    outf.write( "category\tcounts\n%s\n" % counts.asTable() )
    outf.close()

## needs to put into class
##
def deseqPlotSizeFactors(outfile):
    '''plot size factors - requires cds object.'''
    R.png( outfile )
    R('''par(mar=c(8,4,4,2))''')
    R('''barplot( sizeFactors( cds ), main="size factors", las=2)''')
    R['dev.off']()

def deseqOutputSizeFactors( outfile ):
    '''output size factors - requires cds object.'''
    size_factors = R('''sizeFactors( cds )''')
    samples = R('''names(sizeFactors(cds))''')
    with IOTools.openFile( outfile, "w" ) as outf:
        outf.write( "sample\tfactor\n" )
        for name, x in zip( samples, size_factors):
            outf.write( "%s\t%s\n" % (name, str(x)))

def deseqPlotHeatmap( outfile, method = 'pooled', fit_type = 'parametric' ):
    '''plot a heatmap.'''
    
    if method == "per-condition":
        # required to call "pooled" or "blind" if method = per-condition 
        R('''cds <- estimateDispersions( cds, 
                                         method='pooled',
                                         fitType='%(fit_type)s' )''' % locals())
        
    R('''vsd <- getVarianceStabilizedData( cds )''' )
    R('''dists <- dist( t( vsd ) )''')
    R.png( outfile )
    R('''heatmap( as.matrix( dists ), symm=TRUE )''' )
    R['dev.off']()

def deseqPlotPairs( outfile ):
    '''requires counts table'''
    # Plot pairs
    R.png( outfile, width=960, height=960 )
    R('''panel.pearson <- function(x, y, digits=2, prefix="", cex.cor, ...)
            {
            usr <- par("usr"); on.exit(par(usr))
            par(usr = c(0, 1, 0, 1))
            r <- abs(cor(x, y))
            txt <- format(c(r, 0.123456789), digits=digits)[1]
            txt <- paste(prefix, txt, sep="")
            if(missing(cex.cor)) cex <- 0.6/strwidth(txt)
            x = 0.5;
            y = 0.5;
            if (par("xlog")) { x = 10^x }; 
            if (par("ylog")) { y = 10^y }; 
            text(x, y, txt, cex = cex);
            }
       ''')
    R('''pairs( countsTable, lower.panel = panel.pearson, pch=".", log="xy" )''')
    R['dev.off']()

def deseqParseResults( control_name, treatment_name, fdr):
    '''parse deseq output.

    retrieve deseq results from object 'res' in R namespace.

    The 'res' object is a dataframe with the following columns (from the DESeq manual):

    id: The ID of the observable, taken from the row names of the
          counts slots.

    baseMean: The base mean (i.e., mean of the counts divided by the size
          factors) for the counts for both conditions

    baseMeanA: The base mean (i.e., mean of the counts divided by the size
          factors) for the counts for condition A

    baseMeanB: The base mean for condition B

    foldChange: The ratio meanB/meanA

    log2FoldChange: The log2 of the fold change

    pval: The p value for rejecting the null hypothesis 'meanA==meanB'

    padj: The adjusted p values (adjusted with 'p.adjust( pval,
          method="BH")')

    Here, 'conditionA' is 'control' and 'conditionB' is 'treatment' such that
    a foldChange of 2 means that treatment is twice upregulated compared to control.

    Returns a list of results.
    '''

    results = []
    isna = R["is.na"]

    # Get column names from output and edit
    names = list(R['res'].names)
    m = dict( [ (x,x) for x in names ])
    m.update( dict(
            pval = "pvalue", 
            baseMeanA = "value1", 
            baseMeanB = "value2",
            id = "interval_id", 
            log2FoldChange = "lfold") )
    
    rtype = collections.namedtuple( "rtype", names )
    counts = E.Counter()
    
    for data in zip( *R['res']) :
        counts.input += 1

        d = rtype._make( data )

        # set significant flag
        if d.padj <= fdr: 
            signif = 1
            counts.significant += 1
            if d.log2FoldChange > 0:
                counts.significant_over += 1
            else:
                counts.significant_under += 1
        else: 
            signif = 0
            counts.insignificant += 1

        if d.log2FoldChange > 0:
            counts.all_over += 1
        else:
            counts.all_under += 1

        # set lfold change to 0 if both are not expressed
        if d.baseMeanA == 0.0 and d.baseMeanB == 0.0:
            d = d._replace( foldChange = 0, log2FoldChange = 0 )

        if isna( d.pval ): status = "OK"
        else: status = "FAIL"

        counts[status] += 1

        counts.output += 1

        # check if our assumptions about the direction of fold change
        # are correct
        assert (d.foldChange > 1) == (d.baseMeanB > d.baseMeanA )

        # note that fold change is computed as second group (B) divided by first (A)
        results.append( GeneExpressionResult._make( ( \
                    d.id,
                    treatment_name,
                    d.baseMeanB,
                    0,
                    control_name,
                    d.baseMeanA,
                    0,
                    d.pval,
                    d.padj,
                    d.log2FoldChange,
                    d.foldChange,
                    str(signif),
                    status) ) )
                    
    return results, counts

def runDESeq( infile, 
              design_file, 
              outfile, 
              outfile_prefix = "deseq.",
              fdr = 0.1,
              prefix = "",
              fit_type = "parametric",
              dispersion_method = "pooled",
              ref_group = None
              ):
    '''run DESeq on.

    See loadTagData on the input form format for *infile* and
    *design_file*.

    Results are stored in *outfile* and files prefixed by *outfile_prefix*.

    DESeq ignores any pair information in the design matrix.
    
    Various plots are generate - annotation is from the manual (version 1.4)


    SVCPlot:
       squared coefficient of variation. Ratio of variance at base level to the
       square of the base mean.

       Solid lines are for the raw variances (biological noise).

       On top of the variance, there is shot noise, i.e., the Poissonean variance inherent to the
       process of counting reads. The amount of shot noise depends on the size factor, and hence, for
       each sample, a dotted line in the colour of its condition is plotted above the solid line. The dotted
       line is the base variance, i.e., the full variance, scaled down to base level by the size factors. The
       vertical distance between solid and dotted lines is the shot noise.
       The solid black line is a density estimate of the base means: Only were there is an appreciable
       number of base mean values, the variance estimates can be expected to be accurate.
       It is instructive to observe at which count level the biological noise starts to dominate the shot
       noise. At low counts, where shot noise dominates, higher sequencing depth (larger library size)
       will improve the signal-to-noise ratio while for high counts, where the biological noise dominates,
       only additional biological replicates will help.

   fit.png

       One should check whether the base variance functions seem to follow the empirical variance
       well. To this end, two diagnostic functions are provided. The function varianceFitDiagnostics
       returns, for a speci?ed condition, a data frame with four columns: the mean base level for each
       gene, the base variance as estimated from the count values of this gene only, and the ?tted base
       variance, i.e., the predicted value from the local ?t through the base variance estimates from
       all genes. As one typically has few replicates, the single-gene estimate of the base variance can
       deviate wildly from the ?tted value. To see whether this might be too wild, the cumulative prob-
       ability for this ratio of single-gene estimate to ?tted value is calculated from the ?2 distribution,
       as explained in the paper.      

       We may now plot the per-gene estimates of the base variance against the base levels and draw
       a line with the ?t from the local regression

    residuals.png
       Another way to study the diagnostic data is to check whether the probabilities in the fourth
       column of the diagnostics data frame are uniform, as they should be. One may simply look at the
       histogram of diagForGB$pchisq but a more convenient way is the function residualsEcdfPlot,
       which show empirical cumulative density functions (ECDF) strati?ed by base level.

    The output is treatment and control. Fold change values are computed as treatment divided by control.
    '''

    # load library 
    R('''suppressMessages(library('DESeq'))''')

    loadTagData( infile, design_file)

    nobservations, nsamples = filterTagData()

    if nobservations == 0:
        E.warn( "no observations - no output" )
        return

    if nsamples == 0:
        E.warn( "no samples remain after filtering - no output" )
        return

    groups, pairs, has_replicates, has_pairs = groupTagData(ref_group)

    sample_names = R('''colnames(countsTable)''')
    E.info( "%i samples to test at %i observations: %s" % ( nsamples, nobservations,
                                                            ",".join( sample_names)))

    ######## Run DESeq
    # Create Count data object
    E.info( "running DESeq: replicates=%s" % (has_replicates))
    R('''cds <-newCountDataSet( countsTable, groups) ''')

    # Estimate size factors
    R('''cds <- estimateSizeFactors( cds )''')
    
    no_size_factors = R('''is.na(sum(sizeFactors(cds)))''' )[0]
    if no_size_factors:
        E.warn( "no size factors - can not estimate - no output" )
        return

    # Estimate variance
    if has_replicates:
        E.info("replicates - estimating variance from replicates" )
    else:
        E.info("no replicates - estimating variance with method='blind'" )
        dispersion_method = "blind"

    E.info( "Dispersion method = %s, fit type =%s" % (dispersion_method, fit_type ) )
    R('''cds <- estimateDispersions( cds, 
                                     method='%(dispersion_method)s',
                                     fitType='%(fit_type)s' )''' % locals())

    # Plot size factors
    deseqPlotSizeFactors( '%(outfile_prefix)ssize_factors.png''' % locals() )

    # output size factors
    deseqOutputSizeFactors( "%(outfile_prefix)ssize_factors.tsv" % locals() ) 

    deseqPlotPairs('%(outfile_prefix)spairs.png' % locals()) 

    # in DESeq versions > 1.6 the following can be used
    # to output normalized data
    # R('''write.table( counts(cds, normalized=TRUE), file='%(outfile_prefix)scounts.tsv.gz', sep='\t') ''' % locals())
    # output counts
    R('''write.table( counts(cds), file=gzfile('%(outfile_prefix)scounts.tsv.gz'), sep='\t') ''' % locals())
    
    # R.png( '''%(outfile_prefix)sscvplot.png''' % locals() )
    # R('''scvPlot( cds, ylim = c(0,3))''')
    # R['dev.off']()

    deseqPlotHeatmap( '%(outfile_prefix)sheatmap.png' % locals(), 
                      method = dispersion_method, 
                      fit_type = fit_type )

    for group in groups:
        if has_replicates:
            #R.png( '''%(outfile_prefix)s%(group)s_fit.png''' % locals() )
            #R('''diagForT <- varianceFitDiagnostics( cds, "%s" )''' % group )
            #R('''smoothScatter( log10(diagForT$baseMean), log10(diagForT$baseVar) )''')
            #R('''lines( log10(fittedBaseVar) ~ log10(baseMean), diagForT[ order(diagForT$baseMean), ], col="red" )''')
            #R['dev.off']()
            #R.png( '''%(outfile_prefix)s%(group)s_residuals.png''' % locals()  )
            #R('''residualsEcdfPlot( cds, "%s" )''' % group )
            #R['dev.off']()
            pass

    # Call diffential expression for all pairings of groups included in the design
    
    all_results = []
    for combination in itertools.combinations(groups,2):
        control, treatment = combination
        gfix = "%s_vs_%s_" % (control,treatment)

        outfile_groups_prefix = outfile_prefix + gfix
        E.info("calling differential expression for control=%s vs treatment=%s" % (control, treatment))
        R('''res <- nbinomTest( cds, '%s', '%s' )''' % (control,treatment) )

        # Plot significance
        R.png( '''%(outfile_groups_prefix)ssignificance.png''' % locals() )
        R('''plot( res$baseMean, res$log2FoldChange, log="x", pch=20, cex=.1, 
                        col = ifelse( res$padj < %(fdr)s, "red", "black" ) )''' % locals() )
        R['dev.off']()

        outf = IOTools.openFile( "%(outfile_groups_prefix)sall.txt" % locals(), "w" )
        isna = R["is.na"]

        E.info("Generating output (%s vs %s)" % (control, treatment))

        # Parse results and parse to file
        results, counts = deseqParseResults( control, treatment, fdr = fdr )

        all_results += results

        E.info( counts )
        outf = IOTools.openFile( "%(outfile_groups_prefix)ssummary.tsv" % locals(), "w" )
        outf.write( "category\tcounts\n%s\n" % counts.asTable() )
        outf.close()

    if outfile == sys.stdout:
        writeExpressionResults(outfile, all_results)
    else:
        with IOTools.openFile(outfile, "w") as outf:
            writeExpressionResults(outf, all_results)


Design = collections.namedtuple( "Design", ("include", "group", "pair") )

def readDesignFile( design_file ):
    '''reads a design file.'''
    
    design = collections.OrderedDict()
    with IOTools.openFile( design_file ) as inf:
        for line in inf:
            if line.startswith("track"): continue
            track, include, group, pair = line[:-1].split("\t")
            if track in design: raise ValueError( "duplicate track '%s'" % track )
            design[track] = Design._make( (int(include), group, pair))
    return design

#########################################################################
#########################################################################
#########################################################################
def plotTagStats( infile, design_file, outfile ):
    '''provide summary plots for tag data.'''

    loadTagData( infile, design_file )

    nobservations, nsamples = filterTagData()

    if nobservations == 0:
        E.warn( "no observations - no output" )
        return

    if nsamples == 0:
        E.warn( "no samples remain after filtering - no output" )
        return

    groups, pairs, has_replicates, has_pairs = groupTagData() 

    # import rpy2.robjects.lib.ggplot2 as ggplot2

    R('''library('ggplot2')''')
    R('''library('reshape')''')

    R('''d = melt( log10(countsTable + 1), variable_name = 'sample' )''')
    R('''gp = ggplot(d)''')
    R('''pp = gp + \
        geom_density(aes(x=value,group=sample,color=sample,fill=sample),alpha=I(1/3))''')
    
    R.ggsave( outfile + ".densities.png" )
    R['dev.off']()

    R('''gp = ggplot(d)''')
    R('''pp = gp + \
        geom_boxplot(aes(x=sample,y=value,color=sample,fill=sample),size=0.3,alpha=I(1/3)) + 
        opts( axis.text.x = theme_text( angle=90, hjust=1, size=8 ) )''')

    R.ggsave( outfile + ".boxplots.png" )
    R['dev.off']()

#########################################################################
#########################################################################
#########################################################################
def plotDETagStats( infile, outfile ):
    '''provide summary plots for tag data.

    Stratify boxplots and densities according to differential expression calls.
    '''

    # import rpy2.robjects.lib.ggplot2 as ggplot2

    R('''library('ggplot2')''')
    R('''library('grid')''')
    R('''data = read.table( '%s', header = TRUE, row.names=1 )''' % infile ) 

    R(''' gp = ggplot(data)''')
    R('''a = gp + 
        geom_density(aes(x=log10(treatment_mean+1),group=factor(significant),
                                         color='factor(significant)',fill='factor(significant)'),alpha=I(1/3))''')

    R('''b = gp + 
        geom_density(aes(x=log10(control_mean+1),group=factor(significant),
                                         color=factor(significant),fill=factor(significant)),alpha=I(1/3))''')
    

    fn = outfile + ".densities.png" 
    R.png( fn )
    try:
        R('''grid.newpage()''')
        R.pushViewport(R.viewport( layout = R('''grid.layout''')(2,1)))
        R('''print( a, vp = viewport( layout.pos.row = 1, layout.pos.col = 1 ) )''')
        R('''print( b, vp = viewport( layout.pos.row = 2, layout.pos.col = 1 ) )''')
    except rpy2.rinterface.RRuntimeError:
        E.warn( "could not create %s" % fn )
    R['dev.off']()

    
    R('''grid.newpage()''')
    R.pushViewport(R.viewport( layout = R('''grid.layout''')(2,1)))

    R(''' gp = ggplot(data)''')
    R('''a = gp + 
        geom_boxplot(aes(x=factor(significant), y=log10(treatment_mean+1),
                                         color=factor(significant),fill=factor(significant)),
                             size=0.3,
                             alpha=I(1/3))''') 

    R('''b = gp + 
      geom_boxplot(aes(x=factor(significant), 
                       y=log10(control_mean+1),
                         color=factor(significant),
                         fill=factor(significant)),
                         size=0.3,
                         alpha=I(1/3)) +\
        opts( axis.text.x = theme_text( angle=90, hjust=1, size=8 ) )''')

    fn = outfile + ".boxplots.png" 
    R.png( fn )
    try:
        R('''print( a, vp = viewport( layout.pos.row = 1, layout.pos.col = 1 ) )''')
        R('''print( b, vp = viewport( layout.pos.row = 2, layout.pos.col = 1 ) )''')
    except rpy2.rinterface.RRuntimeError:
        E.warn( "could not create %s" % fn )
    R['dev.off']()

def parseCuffdiff( infile):
    '''parse a cuffdiff .diff output file.'''
    min_fpkm = PARAMS["cuffdiff_fpkm_expressed"]

    CuffdiffResult = collections.namedtuple("CuffdiffResult",
                                            "test_id gene_id gene  locus   sample_1 sample_2  " 
                                            " status  value_1 value_2 l2fold  " 
                                            "test_stat p_value q_value significant " )
    
    results = []

    for line in IOTools.openFile( infile ):
        if line.startswith("test_id"): continue
        data = CuffdiffResult._make( line[:-1].split("\t"))
        status = data.status
        significant = [0,1][data.significant == "yes"]
        if status == "OK" and (float(data.value_1) < min_fpkm or float(data.value_2) < min_fpkm):
            status = "NOCALL"

        try: fold = math.pow(2.0, float(data.l2fold))
        except OverflowError: fold = "na"

        results.append( GeneExpressionResult._make( (
                    data.test_id,
                    data.sample_1,
                    data.value_1,
                    0,
                    data.sample_2,
                    data.value_2,
                    0,
                    data.p_value,
                    data.q_value,
                    data.l2fold,
                    fold,
                    significant,
                    status ) ) )
                                            
    return results

#########################################################################
#########################################################################
#########################################################################
def loadCuffdiff( infile, outfile ):
    '''load results from differential expression analysis and produce
    summary plots.

    Note: converts from ln(fold change) to log2 fold change.
   
    The cuffdiff output is parsed. 

    Pairwise comparisons in which one gene is not expressed (fpkm < fpkm_silent)
    are set to status 'NOCALL'. These transcripts might nevertheless be significant.

    This requires the cummeRbund library to be present in R.
    '''

    prefix = P.toTable( outfile )
    indir = infile + ".dir"

    if not os.path.exists( indir ):
        P.touch( outfile )
        return

    # E.info( "building cummeRbund database" )
    # R('''library(cummeRbund)''')
    # cuff = R('''readCufflinks(dir = %(indir)s, dbfile=%(indir)s/csvdb)''' )
    # to be continued

    to_cluster = False
    dbhandle = sqlite3.connect( PARAMS["database"] )

    tmpname = P.getTempFilename()    

    # ignore promoters and splicing - no fold change column, but  sqrt(JS)
    for fn, level in ( ("cds_exp.diff", "cds"),
                       ("gene_exp.diff", "gene"),
                       ("isoform_exp.diff", "isoform"),
                       # ("promoters.diff", "promotor"),
                       # ("splicing.diff", "splice"), 
                       ("tss_group_exp.diff", "tss") ):
        
        tablename = prefix + "_" + level + "_diff"

        infile = os.path.join( indir, fn)                
        results = parseCuffdiff( infile )

        with IOTools.openFile( tmpname, "w" ) as outf:
            writeExpressionResults( outf, results )
            
        statement = '''cat %(tmpname)s 
        | python %(scriptsdir)s/csv2db.py %(csv2db_options)s
              --allow-empty
              --index=treatment_name
              --index=control_name
              --index=test_id
              --table=%(tablename)s 
         >> %(outfile)s.log
         '''
        
        P.run()

    for fn, level in ( ("cds.fpkm_tracking", "cds" ),
                       ("genes.fpkm_tracking", "gene"),
                       ("isoforms.fpkm_tracking", "isoform"),
                       ("tss_groups.fpkm_tracking", "tss") ):

        tablename = prefix + "_" + level + "_levels" 

        statement = '''cat %(indir)s/%(fn)s
        | python %(scriptsdir)s/csv2db.py %(csv2db_options)s
              --allow-empty
              --index=tracking_id
              --table=%(tablename)s 
         >> %(outfile)s.log
         '''
        
        P.run()


    # Jethro - load tables of sample specific cuffdiff fpkm values into csvdb
    for fn, level in ( ("cds.read_group_tracking", "cds" ),
                       ("genes.read_group_tracking", "gene"),
                       ("isoforms.read_group_tracking", "isoform"),
                       ("tss_groups.read_group_tracking", "tss") ):

        tablename = prefix + "_" + level + "sample_fpkms"
        
        tmpf = P.getTempFilename(".")
        inf = IOTools.openFile( os.path.join( indir, fn ) ).readlines()
        outf = IOTools.openFile( tmpf, "w" )

        samples = []
        genes= {}

        x = 0
        for line in inf:
            if x == 0:
                x += 1
                continue
            line = line.split()  
            gene_id = line[0]
            condition = line[1]
            replicate = line[2]
            fpkm = line[6]
            status = line[8]
    
            sample_id = condition + "_" + replicate
            
            if sample_id not in samples:
                samples.append( sample_id )
                
                if gene_id not in genes:
                    genes[gene_id] = {}
                    genes[gene_id][sample_id] = fpkm
                else: 
                    if sample_id in genes[gene_id]:
                        raise ValueError(
                            'sample_id %s appears twice in file for gene_id %s'
                            % ( sample_id, gene_id ) )
                    else:
                        if status != "OK":
                            genes[gene_id][sample_id] = status
                        else:
                            genes[gene_id][sample_id] = fpkm 


        samples = sorted( samples )

        headers = "gene_id\t" + "\t".join( [ x for x in samples ] )
        outf.write( headers + "\n" )

        for gene in genes.iterkeys():
            outf.write( gene + "\t" )
            x = 0
            while x < len( samples ) - 1:
                outf.write( genes[gene][samples[x]] + "\t" )
                x += 1
                outf.write( genes[gene][samples[len(samples)-1]] + "\n" )

        outf.close()


        statement = ( "cat %(tmpf)s |"
                      " python %(scriptsdir)s/csv2db.py "
                      "  %(csv2db_options)s"
                      "  --allow-empty"
                      "  --index=gene_id"
                      "  --table=%(tablename)s"
                      " >> %(outfile)s.log" )
        P.run()

        os.unlink( tmpf )


    ## build convenience table with tracks
    tablename = prefix + "_isoform_levels"
    tracks = Database.getColumnNames( dbhandle, tablename )
    tracks = [ x[:-len("_FPKM")] for x in tracks if x.endswith("_FPKM") ]
    
    tmpfile = P.getTempFile()
    tmpfile.write( "track\n" )
    tmpfile.write("\n".join(tracks) + "\n" )
    tmpfile.close()
    
    statement = P.load( tmpfile.name, outfile )
    os.unlink( tmpfile.name )

def runCuffdiff( bamfiles, 
                 design_file,
                 geneset_file,
                 outfile,
                 cuffdiff_options = "",
                 threads = 4,
                 fdr = 0.1,
                 mask_file = None ):
    '''estimate differential expression using cuffdiff.

    infiles
       bam files

    geneset_file
       geneset to use for the analysis

    design_file
       design file describing which differential expression to test

    Replicates within each track are grouped.
    '''

    design = readDesignFile( design_file )

    to_cluster = True

    outdir = outfile + ".dir" 
    try: os.mkdir( outdir )
    except OSError: pass

    job_options= "-pe dedicated %i -R y" % threads

    # replicates are separated by ","
    reps = collections.defaultdict( list )
    for bamfile in bamfiles:
        groups = collections.defaultdict()
        # .accepted.bam kept for legacy reasons (see rnaseq pipeline)
        track = P.snip( os.path.basename( bamfile ), ".bam", ".accepted.bam" )
        if track not in design:
            E.warn( "bamfile '%s' not part of design - skipped" % bamfile )
            continue
        
        d = design[track]
        if not d.include: continue
        reps[d.group].append( bamfile )
        
    groups = sorted(reps.keys())
    labels = ",".join( groups )
    reps = "   ".join( [ ",".join( reps[group] ) for group in groups ] )

    # Nick - add mask gtf to not assess rRNA and ChrM
    extra_options = []

    if mask_file:
        extra_options.append( " -M %s" % os.path.abspath( mask_file ) )

    extra_options = " ".join( extra_options )

    statement = '''date > %(outfile)s.log; hostname >> %(outfile)s.log;
    cuffdiff --output-dir %(outdir)s
             --verbose
             --num-threads %(threads)i
             --labels %(labels)s
             --FDR %(fdr)f
             %(extra_options)s
             %(cuffdiff_options)s
             <(gunzip < %(geneset_file)s )
             %(reps)s
    >> %(outfile)s.log 2>&1;
    date >> %(outfile)s.log;
    '''
    P.run()

    results = parseCuffdiff( os.path.join( outdir, "gene_exp.diff") )
    
    if outfile == sys.stdout:
        writeExpressionResults( outfile, results )
    else:
        with IOTools.openFile( outfile, "w" ) as outf:
            writeExpressionResults( outf, results )
    
def outputTagSummary( filename_tags, outfile, output_filename_pattern ):
    '''output summary values for a count table.'''

    E.info( "loading tag data from %s" % filename_tags)
    
    outfile.write( "metric\tvalues\n" )

    R( '''countsTable = read.delim( '%(filename_tags)s', 
                                     header = TRUE,
                                     row.names = 1,
                                     stringsAsFactors = TRUE,
                                     comment.char = '#' )''' % locals() )

    nrows, ncolumns = tuple(R('''dim(countsTable)'''))
    E.info( "read data: %i observations for %i samples" % (nrows,ncolumns))
    E.debug( "sample names: %s" % R('''colnames(countsTable)'''))

    outfile.write( "number of rows\t%i\n" % nrows )
    outfile.write( "number of columns\t%i\n" % ncolumns )
    
    # Count windows with no data
    R( '''max_counts = apply(countsTable,1,max)''' )
    
    # output distribution of maximum number of counts per window
    outfilename = output_filename_pattern + "max_counts.tsv.gz"
    E.info( "outputting maximum counts per window to %s" % outfilename )
    R( '''write.table( table(max_counts), file='%(outfilename)s', sep="\t", row.names=FALSE, quote=FALSE)''' % locals())

    # removing empty rows
    E.info( "removing rows with no counts" )
    R( '''countsTable = countsTable[max_counts>0,]''')

    for x in range( 0,20):
        nempty = tuple( R('''sum(max_counts <= %i)''' % x))[0]
        outfile.write( "max<=%i\t%i\t%f\n" % (x, nempty, 100.0 * nempty / nrows ) )
                       
    E.info( "removed %i empty rows" % tuple( R('''sum(max_counts == 0)''') ) )
    observations, samples = tuple( R('''dim(countsTable)'''))
    E.info( "trimmed data: %i observations for %i samples" % (observations, samples ))

    outfilename = output_filename_pattern + "heatmap.svg"
    R.svg( outfilename )
    plotHeatmap()
    R['dev.off']()    



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
csv2xls.py - convert table to excel format
==========================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python csv2xls.py --help

Type::

   python csv2xls.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import getopt
import time
import optparse
import math
import tempfile

import CGAT.Experiment as E
import csv
import CGAT.CSV as CSV

import openpyxl

if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: csv2xls.py 2782 2009-09-10 11:40:29Z andreas $")
    
    parser.add_option( "-o", "--outfile=", dest="output_filename", type="string",
                       help="write to output filename." )

    parser.set_defaults(
        output_filename=None,
        )

    (options, args) = E.Start( parser, add_csv_options  = True)

    if not options.output_filename:
        raise ValueError("please specify an output filename.")

    w = openpyxl.Workbook( optimized_write = True )

    ## create styles
    header_style = GetHeaderStyle()
    data_style = GetDataStyle()

    for filename in args:

        lines = filter( lambda x: x[0] != "#", open(filename, "r").readlines())

        if len(lines) == 0: continue

        if options.loglevel >= 2:
            print "# read %i rows" % len(lines)
            sys.stdout.flush()

        headers = lines[0][:-1].split("\t")

        ws = w.add_sheet(os.path.basename(filename))
        
        cur_row = 0
        
        ws.append( headers )
            
        cur_row += 1
        
        reader = csv.DictReader( lines, dialect=options.csv_dialect )

        for row in reader:
            row = CSV.ConvertDictionary( row )

            data = [ row.get( headers[x], "") for x in range(len(headers))]
            ws.append( data )
                
            cur_row += 1

    
    w.save(options.output_filename)
        
    E.Stop()


